"""Miscellaneous utility functions and classes specific to ansible cli tools."""

from __future__ import annotations

import json
import os
import shutil
import typing as t

from .constants import (
    ANSIBLE_BIN_SYMLINK_MAP,
    SOFT_RLIMIT_NOFILE,
)

from .util import (
    common_environment,
    ApplicationError,
    ANSIBLE_LIB_ROOT,
    ANSIBLE_TEST_ROOT,
    ANSIBLE_TEST_DATA_ROOT,
    ANSIBLE_ROOT,
    ANSIBLE_SOURCE_ROOT,
    ANSIBLE_TEST_TOOLS_ROOT,
    MODE_FILE_EXECUTE,
    raw_command,
    verified_chmod,
)

from .util_common import (
    create_temp_dir,
    ResultType,
    intercept_python,
    get_injector_path,
)

from .config import (
    IntegrationConfig,
    PosixIntegrationConfig,
    EnvironmentConfig,
    CommonConfig,
)

from .data import (
    data_context,
)

from .python_requirements import (
    install_requirements,
)

from .host_configs import (
    PythonConfig,
)

from .thread import (
    mutex,
)


def parse_inventory(args: EnvironmentConfig, inventory_path: str) -> dict[str, t.Any]:
    """Return a dict parsed from the given inventory file."""
    cmd = ['ansible-inventory', '-i', inventory_path, '--list']
    env = ansible_environment(args)
    inventory = json.loads(intercept_python(args, args.controller_python, cmd, env, capture=True, always=True)[0])
    return inventory


def get_hosts(inventory: dict[str, t.Any], group_name: str) -> dict[str, dict[str, t.Any]]:
    """Return a dict of hosts from the specified group in the given inventory."""
    hostvars = inventory.get('_meta', {}).get('hostvars', {})
    group = inventory.get(group_name, {})
    host_names = group.get('hosts', [])
    hosts = dict((name, hostvars.get(name, {})) for name in host_names)
    return hosts


def ansible_environment(args: CommonConfig, color: bool = True, ansible_config: t.Optional[str] = None) -> dict[str, str]:
    """Return a dictionary of environment variables to use when running Ansible commands."""
    env = common_environment()
    path = env['PATH']

    ansible_bin_path = get_ansible_bin_path(args)

    if not path.startswith(ansible_bin_path + os.path.pathsep):
        path = ansible_bin_path + os.path.pathsep + path

    if not ansible_config:
        # use the default empty configuration unless one has been provided
        ansible_config = args.get_ansible_config()

    if not args.explain and not os.path.exists(ansible_config):
        raise ApplicationError('Configuration not found: %s' % ansible_config)

    ansible = dict(
        ANSIBLE_PYTHON_MODULE_RLIMIT_NOFILE=str(SOFT_RLIMIT_NOFILE),
        ANSIBLE_FORCE_COLOR='%s' % 'true' if args.color and color else 'false',
        ANSIBLE_FORCE_HANDLERS='true',  # allow cleanup handlers to run when tests fail
        ANSIBLE_HOST_PATTERN_MISMATCH='error',  # prevent tests from unintentionally passing when hosts are not found
        ANSIBLE_INVENTORY='/dev/null',  # force tests to provide inventory
        ANSIBLE_DEPRECATION_WARNINGS='false',
        ANSIBLE_HOST_KEY_CHECKING='false',
        ANSIBLE_RETRY_FILES_ENABLED='false',
        ANSIBLE_DISPLAY_TRACEBACK=args.display_traceback,
        ANSIBLE_CONFIG=ansible_config,
        ANSIBLE_LIBRARY='/dev/null',
        ANSIBLE_DEVEL_WARNING='false',  # Don't show warnings that CI is running devel
        PYTHONPATH=get_ansible_python_path(args),
        PAGER='/bin/cat',
        PATH=path,
        # give TQM worker processes time to report code coverage results
        # without this the last task in a play may write no coverage file, an empty file, or an incomplete file
        # enabled even when not using code coverage to surface warnings when worker processes do not exit cleanly
        ANSIBLE_WORKER_SHUTDOWN_POLL_COUNT='100',
        ANSIBLE_WORKER_SHUTDOWN_POLL_DELAY='0.1',
        # ansible-test specific environment variables require an 'ANSIBLE_TEST_' prefix to distinguish them from ansible-core env vars defined by config
        ANSIBLE_TEST_ANSIBLE_LIB_ROOT=ANSIBLE_LIB_ROOT,  # used by the coverage injector
    )

    if isinstance(args, IntegrationConfig) and args.coverage:
        # standard path injection is not effective for the persistent connection helper, instead the location must be configured
        # it only requires the injector for code coverage
        # the correct python interpreter is already selected using the sys.executable used to invoke ansible
        ansible.update(
            _ANSIBLE_CONNECTION_PATH=os.path.join(get_injector_path(), 'ansible_connection_cli_stub.py'),
        )

    if isinstance(args, PosixIntegrationConfig):
        ansible.update(
            ANSIBLE_PYTHON_INTERPRETER='/set/ansible_python_interpreter/in/inventory',  # force tests to set ansible_python_interpreter in inventory
        )

    env.update(ansible)

    if args.debug:
        env.update(
            ANSIBLE_DEBUG='true',
            ANSIBLE_LOG_PATH=os.path.join(ResultType.LOGS.name, 'debug.log'),
        )

    if data_context().content.collection:
        env.update(
            ANSIBLE_COLLECTIONS_PATH=data_context().content.collection.root,
        )

    if data_context().content.is_ansible:
        env.update(configure_plugin_paths(args))

    return env


def configure_plugin_paths(args: CommonConfig) -> dict[str, str]:
    """Return environment variables with paths to plugins relevant for the current command."""
    if not isinstance(args, IntegrationConfig):
        return {}

    support_path = os.path.join(ANSIBLE_SOURCE_ROOT, 'test', 'support', args.command)

    # provide private copies of collections for integration tests
    collection_root = os.path.join(support_path, 'collections')

    env = dict(
        ANSIBLE_COLLECTIONS_PATH=collection_root,
    )

    # provide private copies of plugins for integration tests
    plugin_root = os.path.join(support_path, 'plugins')

    plugin_list = [
        'action',
        'become',
        'cache',
        'callback',
        'cliconf',
        'connection',
        'filter',
        'httpapi',
        'inventory',
        'lookup',
        'netconf',
        # 'shell' is not configurable
        'strategy',
        'terminal',
        'test',
        'vars',
    ]

    # most plugins follow a standard naming convention
    plugin_map = dict(('%s_plugins' % name, name) for name in plugin_list)

    # these plugins do not follow the standard naming convention
    plugin_map.update(
        doc_fragment='doc_fragments',
        library='modules',
        module_utils='module_utils',
    )

    env.update(dict(('ANSIBLE_%s' % key.upper(), os.path.join(plugin_root, value)) for key, value in plugin_map.items()))

    # only configure directories which exist
    env = dict((key, value) for key, value in env.items() if os.path.isdir(value))

    return env


@mutex
def get_ansible_bin_path(args: CommonConfig) -> str:
    """
    Return a directory usable for PATH, containing only the ansible entry points.
    If a temporary directory is required, it will be cached for the lifetime of the process and cleaned up at exit.
    """
    try:
        return get_ansible_bin_path.bin_path  # type: ignore[attr-defined]
    except AttributeError:
        pass

    if ANSIBLE_SOURCE_ROOT:
        # when running from source there is no need for a temporary directory since we already have known entry point scripts
        bin_path = os.path.join(ANSIBLE_ROOT, 'bin')
    else:
        # when not running from source the installed entry points cannot be relied upon
        # doing so would require using the interpreter specified by those entry points, which conflicts with using our interpreter and injector
        # instead a temporary directory is created which contains only ansible entry points
        # symbolic links cannot be used since the files are likely not executable
        bin_path = create_temp_dir(prefix='ansible-test-', suffix='-bin')
        bin_links = {os.path.join(bin_path, name): get_cli_path(path) for name, path in ANSIBLE_BIN_SYMLINK_MAP.items()}

        if not args.explain:
            for dst, src in bin_links.items():
                shutil.copy(src, dst)
                verified_chmod(dst, MODE_FILE_EXECUTE)

    get_ansible_bin_path.bin_path = bin_path  # type: ignore[attr-defined]

    return bin_path


def get_cli_path(path: str) -> str:
    """Return the absolute path to the CLI script from the given path which is relative to the `bin` directory of the original source tree layout."""
    path_rewrite = {
        '../lib/ansible/': ANSIBLE_LIB_ROOT,
        '../test/lib/ansible_test/': ANSIBLE_TEST_ROOT,
    }

    for prefix, destination in path_rewrite.items():
        if path.startswith(prefix):
            return os.path.join(destination, path[len(prefix):])

    raise RuntimeError(path)


# noinspection PyUnusedLocal
@mutex
def get_ansible_python_path(args: CommonConfig) -> str:
    """
    Return a directory usable for PYTHONPATH, containing only the ansible package.
    If a temporary directory is required, it will be cached for the lifetime of the process and cleaned up at exit.
    """
    del args  # not currently used

    try:
        return get_ansible_python_path.python_path  # type: ignore[attr-defined]
    except AttributeError:
        pass

    if ANSIBLE_SOURCE_ROOT:
        # when running from source there is no need for a temporary directory to isolate the ansible package
        python_path = os.path.dirname(ANSIBLE_LIB_ROOT)
    else:
        # when not running from source the installed directory is unsafe to add to PYTHONPATH
        # doing so would expose many unwanted packages on sys.path
        # instead a temporary directory is created which contains only ansible using a symlink
        python_path = create_temp_dir(prefix='ansible-test-')

        os.symlink(ANSIBLE_LIB_ROOT, os.path.join(python_path, 'ansible'))

    get_ansible_python_path.python_path = python_path  # type: ignore[attr-defined]

    return python_path


class CollectionDetail:
    """Collection detail."""

    def __init__(self) -> None:
        self.version: t.Optional[str] = None


class CollectionDetailError(ApplicationError):
    """An error occurred retrieving collection detail."""

    def __init__(self, reason: str) -> None:
        super().__init__('Error collecting collection detail: %s' % reason)
        self.reason = reason


def get_collection_detail(python: PythonConfig) -> CollectionDetail:
    """Return collection detail."""
    collection = data_context().content.collection
    directory = os.path.join(collection.root, collection.directory)

    stdout = raw_command([python.path, os.path.join(ANSIBLE_TEST_TOOLS_ROOT, 'collection_detail.py'), directory], capture=True)[0]
    result = json.loads(stdout)
    error = result.get('error')

    if error:
        raise CollectionDetailError(error)

    version = result.get('version')

    detail = CollectionDetail()
    detail.version = str(version) if version is not None else None

    return detail


def run_playbook(
    args: EnvironmentConfig,
    inventory_path: str,
    playbook: str,
    capture: bool,
    variables: t.Optional[dict[str, t.Any]] = None,
) -> None:
    """Run the specified playbook using the given inventory file and playbook variables."""
    playbook_path = os.path.join(ANSIBLE_TEST_DATA_ROOT, 'playbooks', playbook)
    cmd = ['ansible-playbook', '-i', inventory_path, playbook_path]

    if variables:
        cmd.extend(['-e', json.dumps(variables)])

    if args.verbosity:
        cmd.append('-%s' % ('v' * args.verbosity))

    install_requirements(args, None, args.controller_python, ansible=True)  # run_playbook()
    env = ansible_environment(args)
    intercept_python(args, args.controller_python, cmd, env, capture=capture)



from __future__ import annotations

import argparse
import json
import os
import typing as t

import pytest
import pytest_mock

if t.TYPE_CHECKING:
    from ansible_test._internal.ci.azp import AzurePipelinesChanges


def create_azure_pipelines_changes(mocker: pytest_mock.MockerFixture) -> AzurePipelinesChanges:
    """Prepare an AzurePipelinesChanges instance for testing."""
    from ansible_test._internal.ci.azp import AzurePipelinesChanges
    from ansible_test._internal.config import CommonConfig

    namespace = argparse.Namespace()
    namespace.color = False
    namespace.explain = False
    namespace.verbosity = False
    namespace.debug = False
    namespace.truncate = False
    namespace.redact = False
    namespace.display_traceback = False

    config = CommonConfig(namespace, 'sanity')

    env = dict(
        HOME=os.environ['HOME'],
        SYSTEM_COLLECTIONURI='https://dev.azure.com/ansible/',
        SYSTEM_TEAMPROJECT='ansible',
        BUILD_REPOSITORY_PROVIDER='GitHub',
        BUILD_SOURCEBRANCH='devel',
        BUILD_SOURCEBRANCHNAME='devel',
    )

    mocker.patch.dict(os.environ, env, clear=True)

    return AzurePipelinesChanges(config)


@pytest.mark.parametrize("status_code,response,expected_commits,expected_warning", (
    # valid 200 responses
    (200, dict(value=[]), None, None),
    (200, dict(value=[dict(sourceVersion='abc')]), {'abc'}, None),
    # invalid 200 responses
    (200, 'not-json', None, "Unable to find project due to HTTP 200 Non-JSON result."),
    (200, '"not-a-dict"', None, "Unexpected response format from HTTP 200 JSON result: string indices must be integers, not 'str'"),
    (200, dict(value='not-a-list'), None, "Unexpected response format from HTTP 200 JSON result: string indices must be integers, not 'str'"),
    (200, dict(value=['not-a-dict']), None, "Unexpected response format from HTTP 200 JSON result: string indices must be integers, not 'str'"),
    (200, dict(), None, "Missing 'value' key in response from HTTP 200 JSON result."),
    (200, dict(value=[{}]), None, "Missing 'sourceVersion' key in response from HTTP 200 JSON result."),
    # non-200 responses
    (404, '', None, "Unable to find project due to HTTP 404 Non-JSON result."),
    (404, '""', None, "Unable to find project due to HTTP 404 JSON result."),
    (404, dict(value=[]), None, "Unable to find project due to HTTP 404 JSON result."),
))
def test_get_successful_merge_run_commits(
    status_code: int,
    response: object,
    expected_commits: set[str] | None,
    expected_warning: str | None,
    mocker: pytest_mock.MockerFixture,
) -> None:
    """Verify AZP commit retrieval handles invalid responses gracefully."""
    from ansible_test._internal.ci.azp import AzurePipelinesChanges
    from ansible_test._internal.git import Git
    from ansible_test._internal.http import HttpClient, HttpResponse
    from ansible_test._internal.util import display

    if not isinstance(response, str):
        response = json.dumps(response)

    if expected_warning:
        expected_warning = f'Cannot determine changes. All tests will be executed. Reason: {expected_warning}'

    patched_get = mocker.patch.object(HttpClient, 'get', return_value=HttpResponse('GET', 'URL', status_code, response))
    patched_warning = mocker.patch.object(display, 'warning')

    mocker.patch.object(Git, 'run_git', return_value='')  # avoid git

    spy_get_successful_merge_run_commits = mocker.spy(AzurePipelinesChanges, 'get_successful_merge_run_commits')

    create_azure_pipelines_changes(mocker)

    assert patched_get.call_count == 1

    if expected_warning:
        patched_warning.assert_called_once_with(expected_warning)
    else:
        patched_warning.assert_not_called()

    assert spy_get_successful_merge_run_commits.spy_return == (expected_commits or set())



from __future__ import annotations

import dataclasses
import json

import typing as t

from ansible.module_utils._internal._ansiballz import _extensions
from ansible.module_utils._internal._ansiballz._extensions import _debugpy, _pydevd, _coverage
from ansible.constants import config

_T = t.TypeVar('_T')


class ExtensionManager:
    """AnsiballZ extension manager."""

    def __init__(
        self,
        pydevd: _pydevd.Options | None = None,
        debugpy: _debugpy.Options | None = None,
        coverage: _coverage.Options | None = None,
    ) -> None:
        options = dict(
            _pydevd=pydevd,
            _debugpy=debugpy,
            _coverage=coverage,
        )

        self._pydevd = pydevd
        self._debugpy = debugpy
        self._coverage = coverage
        self._extension_names = tuple(name for name, option in options.items() if option)
        self._module_names = tuple(f'{_extensions.__name__}.{name}' for name in self._extension_names)

        self.source_mapping: dict[str, str] = {}

    @property
    def debugger_enabled(self) -> bool:
        """Returns True if the debugger extension is enabled, otherwise False."""
        return bool(self._pydevd or self._debugpy)

    @property
    def extension_names(self) -> tuple[str, ...]:
        """Names of extensions to include in the AnsiballZ payload."""
        return self._extension_names

    @property
    def module_names(self) -> tuple[str, ...]:
        """Python module names of extensions to include in the AnsiballZ payload."""
        return self._module_names

    def get_extensions(self) -> dict[str, dict[str, object]]:
        """Return the configured extensions and their options."""
        extension_options: dict[str, t.Any] = {}

        if self._debugpy:
            extension_options['_debugpy'] = dataclasses.replace(
                self._debugpy,
                source_mapping=self._get_source_mapping(self._debugpy.source_mapping),
            )

        if self._pydevd:
            extension_options['_pydevd'] = dataclasses.replace(
                self._pydevd,
                source_mapping=self._get_source_mapping(self._pydevd.source_mapping),
            )

        if self._coverage:
            extension_options['_coverage'] = self._coverage

        extensions = {extension: dataclasses.asdict(options) for extension, options in extension_options.items()}

        return extensions

    def _get_source_mapping(self, debugger_mapping: dict[str, str]) -> dict[str, str]:
        """Get the source mapping, adjusting the source root as needed."""
        if debugger_mapping:
            source_mapping = {self._translate_path(key, debugger_mapping): value for key, value in self.source_mapping.items()}
        else:
            source_mapping = self.source_mapping

        return source_mapping

    @staticmethod
    def _translate_path(path: str, debugger_mapping: dict[str, str]) -> str:
        """Translate a local path to a foreign path."""
        for replace, match in debugger_mapping.items():
            if path.startswith(match):
                return replace + path[len(match) :]

        return path

    @classmethod
    def create(cls, task_vars: dict[str, object]) -> t.Self:
        """Create an instance using the provided task vars."""
        return cls(
            pydevd=cls._get_options('_ANSIBALLZ_PYDEVD_CONFIG', _pydevd.Options, task_vars),
            debugpy=cls._get_options('_ANSIBALLZ_DEBUGPY_CONFIG', _debugpy.Options, task_vars),
            coverage=cls._get_options('_ANSIBALLZ_COVERAGE_CONFIG', _coverage.Options, task_vars),
        )

    @classmethod
    def _get_options(cls, name: str, config_type: type[_T], task_vars: dict[str, object]) -> _T | None:
        """Parse configuration from the named environment variable as the specified type, or None if not configured."""
        if (value := config.get_config_value(name, variables=task_vars)) is None:
            return None

        data = json.loads(value) if isinstance(value, str) else value
        options = config_type(**data)

        return options



from __future__ import annotations

from ansible.module_utils._internal._datatag import AnsibleTagHelper


def str_problematic_strip(value: str) -> str:
    """
    Return a copy of `value` with leading and trailing whitespace removed.
    Used where `str.strip` is needed, but tags must be preserved *AND* the stripping behavior likely shouldn't exist.
    If the stripping behavior is non-problematic, use `AnsibleTagHelper.tag_copy` around `str.strip` instead.
    """
    if (stripped_value := value.strip()) == value:
        return value

    # FUTURE: consider deprecating some/all usages of this method; they generally imply a code smell or pattern we shouldn't be supporting

    stripped_value = AnsibleTagHelper.tag_copy(value, stripped_value)

    return stripped_value



from __future__ import annotations

import json

from ansible.module_utils import _internal
from ansible.module_utils.common.json import AnsibleJSONEncoder
from ansible.module_utils._internal._json._legacy_encoder import LegacyTargetJSONEncoder


def test_controller_impl_patch() -> None:
    """Ensure that the context-appropriate implementation of `AnsibleJSONEncoder` is patched into the module."""
    if _internal.is_controller:
        assert AnsibleJSONEncoder is not LegacyTargetJSONEncoder
    else:
        assert AnsibleJSONEncoder is LegacyTargetJSONEncoder


def test_mu_bytes() -> None:
    assert json.dumps(b'abc', cls=LegacyTargetJSONEncoder, _decode_bytes=True) == '"abc"'



# (c) 2018 Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import annotations

import base64
import dataclasses
import errno
import json
import os
import pkgutil
import secrets
import re
import typing as t

from importlib import import_module

from ansible.module_utils.compat.version import LooseVersion

from ansible import constants as C
from ansible.module_utils.common.json import Direction, get_module_encoder
from ansible.errors import AnsibleError, AnsibleFileNotFound
from ansible.module_utils.common.text.converters import to_bytes, to_text
from ansible.plugins.become import BecomeBase
from ansible.plugins.become.runas import BecomeModule as RunasBecomeModule
from ansible.plugins.loader import ps_module_utils_loader


@dataclasses.dataclass(frozen=True)
class _ExecManifest:
    scripts: dict[str, _ScriptInfo] = dataclasses.field(default_factory=dict)
    actions: list[_ManifestAction] = dataclasses.field(default_factory=list)
    signed_hashlist: list[str] = dataclasses.field(default_factory=list)


@dataclasses.dataclass(frozen=True, kw_only=True)
class _ScriptInfo:
    content: dataclasses.InitVar[bytes]
    path: str
    script: str = dataclasses.field(init=False)

    def __post_init__(self, content: bytes) -> None:
        object.__setattr__(self, 'script', base64.b64encode(content).decode())


@dataclasses.dataclass(frozen=True, kw_only=True)
class _ManifestAction:
    name: str
    params: dict[str, object] = dataclasses.field(default_factory=dict)
    secure_params: dict[str, object] = dataclasses.field(default_factory=dict)


class PSModuleDepFinder(object):

    def __init__(self) -> None:
        # This is also used by validate-modules to get a module's required utils in base and a collection.
        self.scripts: dict[str, _ScriptInfo] = {}
        self.signed_hashlist: set[str] = set()

        if builtin_hashlist := _get_powershell_signed_hashlist():
            self.signed_hashlist.add(builtin_hashlist.path)
            self.scripts[builtin_hashlist.path] = builtin_hashlist

        self._util_deps: dict[str, set[str]] = {}

        self.ps_version: str | None = None
        self.os_version: str | None = None
        self.become = False

        self._re_cs_module = [
            # Reference C# module_util in another C# util, this must always be the fully qualified name.
            # 'using ansible_collections.{namespace}.{collection}.plugins.module_utils.{name}'
            re.compile(to_bytes(r'(?i)^using\s((Ansible\..+)|'
                                r'(ansible_collections\.\w+\.\w+\.plugins\.module_utils\.[\w\.]+));\s*$')),
        ]

        self._re_cs_in_ps_module = [
            # Reference C# module_util in a PowerShell module
            # '#AnsibleRequires -CSharpUtil Ansible.{name}'
            # '#AnsibleRequires -CSharpUtil ansible_collections.{namespace}.{collection}.plugins.module_utils.{name}'
            # '#AnsibleRequires -CSharpUtil ..module_utils.{name}'
            # Can have '-Optional' at the end to denote the util is optional
            re.compile(to_bytes(r'(?i)^#\s*ansiblerequires\s+-csharputil\s+((Ansible\.[\w\.]+)|'
                                r'(ansible_collections\.\w+\.\w+\.plugins\.module_utils\.[\w\.]+)|'
                                r'(\.[\w\.]+))(?P<optional>\s+-Optional){0,1}')),
        ]

        self._re_ps_module = [
            # Original way of referencing a builtin module_util
            # '#Requires -Module Ansible.ModuleUtils.{name}
            re.compile(to_bytes(r'(?i)^#\s*requires\s+\-module(?:s?)\s*(Ansible\.ModuleUtils\..+)')),
            # New way of referencing a builtin and collection module_util
            # '#AnsibleRequires -PowerShell Ansible.ModuleUtils.{name}'
            # '#AnsibleRequires -PowerShell ansible_collections.{namespace}.{collection}.plugins.module_utils.{name}'
            # '#AnsibleRequires -PowerShell ..module_utils.{name}'
            # Can have '-Optional' at the end to denote the util is optional
            re.compile(to_bytes(r'(?i)^#\s*ansiblerequires\s+-powershell\s+((Ansible\.ModuleUtils\.[\w\.]+)|'
                                r'(ansible_collections\.\w+\.\w+\.plugins\.module_utils\.[\w\.]+)|'
                                r'(\.[\w\.]+))(?P<optional>\s+-Optional){0,1}')),
        ]

        self._re_ps_version = re.compile(to_bytes(r'(?i)^#requires\s+\-version\s+([0-9]+(\.[0-9]+){0,3})$'))
        self._re_os_version = re.compile(to_bytes(r'(?i)^#ansiblerequires\s+\-osversion\s+([0-9]+(\.[0-9]+){0,3})$'))
        self._re_become = re.compile(to_bytes(r'(?i)^#ansiblerequires\s+\-become$'))

    def scan_exec_script(self, name: str) -> None:
        # scans lib/ansible/executor/powershell for scripts used in the module
        # exec side. It also scans these scripts for any dependencies
        if name in self.scripts:
            return

        exec_code = _get_powershell_script(name)
        self.scripts[name] = _ScriptInfo(
            content=exec_code,
            path=name,
        )
        self.scan_module(exec_code, powershell=True)

    def scan_module(
        self,
        module_data: bytes,
        fqn: str | None = None,
        powershell: bool = True,
    ) -> set[str]:
        lines = module_data.split(b'\n')
        module_utils: set[tuple[str, str, bool]] = set()

        if fqn and fqn.startswith("ansible_collections."):
            submodules = fqn.split('.')
            collection_name = '.'.join(submodules[:3])

            collection_hashlist = _get_powershell_signed_hashlist(collection_name)
            if collection_hashlist and collection_hashlist.path not in self.signed_hashlist:
                self.signed_hashlist.add(collection_hashlist.path)
                self.scripts[collection_hashlist.path] = collection_hashlist

        if powershell:
            checks = [
                # PS module contains '#Requires -Module Ansible.ModuleUtils.*'
                # PS module contains '#AnsibleRequires -Powershell Ansible.*' (or collections module_utils ref)
                (self._re_ps_module, ".psm1"),
                # PS module contains '#AnsibleRequires -CSharpUtil Ansible.*' (or collections module_utils ref)
                (self._re_cs_in_ps_module, ".cs"),
            ]
        else:
            checks = [
                # CS module contains 'using Ansible.*;' or 'using ansible_collections.ns.coll.plugins.module_utils.*;'
                (self._re_cs_module, ".cs"),
            ]

        for line in lines:
            for patterns, util_extension in checks:
                for pattern in patterns:
                    match = pattern.match(line)
                    if match:
                        # tolerate windows line endings by stripping any remaining
                        # newline chars
                        module_util_name = to_text(match.group(1).rstrip())
                        match_dict = match.groupdict()
                        optional = match_dict.get('optional', None) is not None
                        module_utils.add((module_util_name, util_extension, optional))
                        break

            if not powershell:
                continue

            if ps_version_match := self._re_ps_version.match(line):
                self._parse_version_match(ps_version_match, "ps_version")

            if os_version_match := self._re_os_version.match(line):
                self._parse_version_match(os_version_match, "os_version")

            # once become is set, no need to keep on checking recursively
            if not self.become and self._re_become.match(line):
                self.become = True

        dependencies: set[str] = set()
        for name, ext, optional in set(module_utils):
            util_name = self._scan_module_util(name, ext, fqn, optional)
            if util_name:
                dependencies.add(util_name)
                util_deps = self._util_deps[util_name]
                dependencies.update(util_deps)

        return dependencies

    def _scan_module_util(
        self,
        name: str,
        ext: str,
        module_fqn: str | None,
        optional: bool,
    ) -> str | None:
        util_name: str
        util_path: str
        util_data: bytes
        util_fqn: str | None = None

        if name.startswith("Ansible."):
            # Builtin util, or the old role module_utils reference.
            util_name = f"{name}{ext}"

            if util_name in self._util_deps:
                return util_name

            util_path = ps_module_utils_loader.find_plugin(name, ext)
            if not util_path or not os.path.exists(util_path):
                if optional:
                    return None

                raise AnsibleError(f"Could not find imported module util '{name}'")

            with open(util_path, 'rb') as mu_file:
                util_data = mu_file.read()

        else:
            # Collection util, load the package data based on the util import.
            submodules = name.split(".")
            if name.startswith('.'):
                fqn_submodules = (module_fqn or "").split('.')
                for submodule in submodules:
                    if submodule:
                        break
                    del fqn_submodules[-1]

                submodules = fqn_submodules + [s for s in submodules if s]

            util_package = '.'.join(submodules[:-1])
            util_resource_name = f"{submodules[-1]}{ext}"
            util_fqn = f"{util_package}.{submodules[-1]}"
            util_name = f"{util_package}.{util_resource_name}"

            if util_name in self._util_deps:
                return util_name

            try:
                module_util = import_module(util_package)
                util_code = pkgutil.get_data(util_package, util_resource_name)
                if util_code is None:
                    raise ImportError("No package data found")
                util_data = util_code

                # Get the path of the util which is required for coverage collection.
                resource_paths = list(module_util.__path__)
                if len(resource_paths) != 1:
                    # This should never happen with a collection but we are just being defensive about it.
                    raise AnsibleError(f"Internal error: Referenced module_util package '{util_package}' contains 0 "
                                       "or multiple import locations when we only expect 1.")

                util_path = os.path.join(resource_paths[0], util_resource_name)
            except (ImportError, OSError) as err:
                if getattr(err, "errno", errno.ENOENT) == errno.ENOENT:
                    if optional:
                        return None

                    raise AnsibleError(f"Could not find collection imported module support code for '{name}'")

                else:
                    raise

        # This is important to be set before scan_module is called to avoid
        # recursive dependencies.
        self.scripts[util_name] = _ScriptInfo(
            content=util_data,
            path=util_path,
        )

        # It is important this is set before calling scan_module to ensure
        # recursive dependencies don't result in an infinite loop.
        dependencies = self._util_deps[util_name] = set()

        util_deps = self.scan_module(util_data, fqn=util_fqn, powershell=(ext == ".psm1"))
        dependencies.update(util_deps)
        for dep in dependencies:
            if dep_list := self._util_deps.get(dep):
                dependencies.update(dep_list)

        if ext == ".cs":
            # Any C# code requires the AddType.psm1 module to load.
            dependencies.add("Ansible.ModuleUtils.AddType.psm1")
            self._scan_module_util("Ansible.ModuleUtils.AddType", ".psm1", None, False)

        return util_name

    def _parse_version_match(self, match: re.Match, attribute: str) -> None:
        new_version = to_text(match.group(1)).rstrip()

        # PowerShell cannot cast a string of "1" to Version, it must have at
        # least the major.minor for it to be valid so we append 0
        if match.group(2) is None:
            new_version = f"{new_version}.0"

        existing_version = getattr(self, attribute, None)
        if existing_version is None:
            setattr(self, attribute, new_version)
        else:
            # determine which is the latest version and set that
            if LooseVersion(new_version) > LooseVersion(existing_version):
                setattr(self, attribute, new_version)


def _bootstrap_powershell_script(
    name: str,
    parameters: dict[str, t.Any] | None = None,
    *,
    has_input: bool = False,
) -> tuple[str, bytes]:
    """Build bootstrap wrapper for specified script.

    Builds the bootstrap wrapper and input needed to run the specified executor
    PowerShell script specified.

    :param name: The name of the PowerShell script to run.
    :param parameters: The parameters to pass to the script.
    :param has_input: The script will be provided with input data.
    :return: The bootstrap wrapper and input to provide to it.
    """
    exec_manifest = _ExecManifest()

    script = _get_powershell_script(name)
    exec_manifest.scripts[name] = _ScriptInfo(
        content=script,
        path=name,
    )

    exec_manifest.actions.append(
        _ManifestAction(
            name=name,
            params=parameters or {},
        )
    )

    if hashlist := _get_powershell_signed_hashlist():
        exec_manifest.signed_hashlist.append(hashlist.path)
        exec_manifest.scripts[hashlist.path] = hashlist

    bootstrap_wrapper = _get_powershell_script("bootstrap_wrapper.ps1")
    bootstrap_input = _get_bootstrap_input(exec_manifest)
    if has_input:
        bootstrap_input += b"\n\0\0\0\0\n"

    return bootstrap_wrapper.decode(), bootstrap_input


def _get_powershell_script(
    name: str,
) -> bytes:
    """Get the requested PowerShell script.

    Gets the script stored in the ansible.executore.powershell package.

    :param name: The name of the PowerShell script to retrieve.
    :return: The contents of the requested PowerShell script as a byte string.
    """
    package_name = 'ansible.executor.powershell'

    code = pkgutil.get_data(package_name, name)
    if code is None:
        raise AnsibleFileNotFound(f"Could not find powershell script '{package_name}.{name}'")

    try:
        sig_data = pkgutil.get_data(package_name, f"{name}.authenticode")
    except FileNotFoundError:
        sig_data = None

    if sig_data:
        code = code + b"\r\n" + b"\r\n".join(sig_data.splitlines()) + b"\r\n"

    return code


def _create_powershell_wrapper(
    *,
    name: str,
    module_data: bytes,
    module_path: str,
    module_args: dict[t.Any, t.Any],
    environment: dict[str, str],
    async_timeout: int,
    become_plugin: BecomeBase | None,
    substyle: t.Literal["powershell", "script"],
    task_vars: dict[str, t.Any],
    profile: str,
) -> bytes:
    """Creates module or script wrapper for PowerShell.

    Creates the input data to provide to bootstrap_wrapper.ps1 when running a
    PowerShell module or script.

    :param name: The fully qualified name of the module or script filename (without extension).
    :param module_data: The data of the module or script.
    :param module_path: The path of the module or script.
    :param module_args: The arguments to pass to the module or script.
    :param environment: The environment variables to set when running the module or script.
    :param async_timeout: The timeout to use for async execution or 0 for no async.
    :param become_plugin: The become plugin to use for privilege escalation or None for no become.
    :param substyle: The substyle of the module or script to run [powershell or script].
    :param task_vars: The task variables used on the task.

    :return: The input data for bootstrap_wrapper.ps1 as a byte string.
    """

    actions: list[_ManifestAction] = []
    finder = PSModuleDepFinder()
    finder.scan_exec_script('module_wrapper.ps1')

    ext = os.path.splitext(module_path)[1]
    name_with_ext = f"{name}{ext}"
    finder.scripts[name_with_ext] = _ScriptInfo(
        content=module_data,
        path=module_path,
    )

    module_params: dict[str, t.Any] = {
        'Script': name_with_ext,
        'Environment': environment,
    }
    if substyle != 'script':
        module_deps = finder.scan_module(
            module_data,
            fqn=name,
            powershell=True,
        )
        cs_deps = []
        ps_deps = []
        for dep in module_deps:
            if dep.endswith('.cs'):
                cs_deps.append(dep)
            else:
                ps_deps.append(dep)

        module_params |= {
            'Variables': [
                {
                    'Name': 'complex_args',
                    'Value': _prepare_module_args(module_args, profile),
                    'Scope': 'Global',
                },
            ],
            'CSharpModules': cs_deps,
            'PowerShellModules': ps_deps,
            'ForModule': True,
        }

    if become_plugin or finder.become:
        become_script = 'become_wrapper.ps1'
        become_params: dict[str, t.Any] = {
            'BecomeUser': 'SYSTEM',
        }
        become_secure_params: dict[str, t.Any] = {}

        if become_plugin:
            if not isinstance(become_plugin, RunasBecomeModule):
                msg = f"Become plugin {become_plugin.name} is not supported by the Windows exec wrapper. Make sure to set the become method to runas."
                raise AnsibleError(msg)

            become_script, become_params, become_secure_params = become_plugin._build_powershell_wrapper_action()

        finder.scan_exec_script('exec_wrapper.ps1')
        finder.scan_exec_script(become_script)
        actions.append(
            _ManifestAction(
                name=become_script,
                params=become_params,
                secure_params=become_secure_params,
            )
        )

    if async_timeout > 0:
        finder.scan_exec_script('bootstrap_wrapper.ps1')
        finder.scan_exec_script('exec_wrapper.ps1')

        async_dir = environment.get('ANSIBLE_ASYNC_DIR', None)
        if not async_dir:
            raise AnsibleError("The environment variable 'ANSIBLE_ASYNC_DIR' is not set.")

        finder.scan_exec_script('async_wrapper.ps1')
        actions.append(
            _ManifestAction(
                name='async_wrapper.ps1',
                params={
                    'AsyncDir': async_dir,
                    'AsyncJid': f'j{secrets.randbelow(999999999999)}',
                    'StartupTimeout': C.config.get_config_value("WIN_ASYNC_STARTUP_TIMEOUT", variables=task_vars),
                },
            )
        )

        finder.scan_exec_script('async_watchdog.ps1')
        actions.append(
            _ManifestAction(
                name='async_watchdog.ps1',
                params={
                    'Timeout': async_timeout,
                },
            )
        )

    coverage_output = C.config.get_config_value('COVERAGE_REMOTE_OUTPUT', variables=task_vars)
    if coverage_output and substyle == 'powershell':
        path_filter = C.config.get_config_value('COVERAGE_REMOTE_PATHS', variables=task_vars)

        finder.scan_exec_script('coverage_wrapper.ps1')
        actions.append(
            _ManifestAction(
                name='coverage_wrapper.ps1',
                params={
                    'ModuleName': name_with_ext,
                    'OutputPath': coverage_output,
                    'PathFilter': path_filter,
                },
            )
        )

    actions.append(
        _ManifestAction(
            name='module_wrapper.ps1',
            params=module_params,
        ),
    )

    temp_path: str | None = None
    for temp_key in ['_ansible_tmpdir', '_ansible_remote_tmp']:
        if temp_value := module_args.get(temp_key, None):
            temp_path = temp_value
            break

    exec_manifest = _ExecManifest(
        scripts=finder.scripts,
        actions=actions,
        signed_hashlist=list(finder.signed_hashlist),
    )

    return _get_bootstrap_input(
        exec_manifest,
        min_os_version=finder.os_version,
        min_ps_version=finder.ps_version,
        temp_path=temp_path,
    )


def _get_bootstrap_input(
    manifest: _ExecManifest,
    min_os_version: str | None = None,
    min_ps_version: str | None = None,
    temp_path: str | None = None,
) -> bytes:
    """Gets the input for bootstrap_wrapper.ps1

    Gets the input needed to send to bootstrap_wrapper.ps1 to run code through
    exec_wrapper.ps1.

    :param manifest: The exec wrapper manifest of scripts and actions to run.
    :param min_os_version: The minimum OS version required to run the scripts.
    :param min_ps_version: The minimum PowerShell version required to run the scripts.
    :param temp_path: The temporary path to use for the scripts if needed.
    :return: The input for bootstrap_wrapper.ps1 as a byte string.
    """
    bootstrap_manifest = {
        'name': 'exec_wrapper',
        'script': _get_powershell_script("exec_wrapper.ps1").decode(),
        'params': {
            'MinOSVersion': min_os_version,
            'MinPSVersion': min_ps_version,
            'TempPath': temp_path,
        },
    }

    bootstrap_input = json.dumps(bootstrap_manifest, ensure_ascii=True)
    exec_input = json.dumps(dataclasses.asdict(manifest))
    return f"{bootstrap_input}\n\0\0\0\0\n{exec_input}".encode()


def _prepare_module_args(module_args: dict[str, t.Any], profile: str) -> dict[str, t.Any]:
    """
    Serialize the module args with the specified profile and deserialize them with the Python built-in JSON decoder.
    This is used to facilitate serializing module args with a different encoder (profile) than is used for the manifest.
    """
    encoder = get_module_encoder(profile, Direction.CONTROLLER_TO_MODULE)

    return json.loads(json.dumps(module_args, cls=encoder))


def _get_powershell_signed_hashlist(
    collection: str | None = None,
) -> _ScriptInfo | None:
    """Gets the signed hashlist script stored in either the Ansible package or for
    the collection specified.

    :param collection: The collection namespace to get the signed hashlist for or None for the builtin.
    :return: The _ScriptInfo payload of the signed hashlist script if found, None if not.
    """
    resource = 'ansible.config' if collection is None else f"{collection}.meta"
    signature_file = 'powershell_signatures.psd1'

    try:
        sig_data = pkgutil.get_data(resource, signature_file)
    except FileNotFoundError:
        sig_data = None

    if sig_data:
        resource_path = f"{resource}.{signature_file}"
        return _ScriptInfo(content=sig_data, path=resource_path)

    return None



from __future__ import annotations

import abc
import collections.abc as c
import copy
import dataclasses
import datetime
import enum
import inspect
import sys

from itertools import chain

# deprecated: description='typing.Self exists in Python 3.11+' python_version='3.10'
from ansible.module_utils.compat import typing as t

from ansible.module_utils._internal import _dataclass_validation
from ansible.module_utils._internal._patches import _sys_intern_patch, _socket_patch

_sys_intern_patch.SysInternPatch.patch()
_socket_patch.GetAddrInfoPatch.patch()  # DTFIX-FUTURE: consider replacing this with a socket import shim that installs the patch

if sys.version_info >= (3, 10):
    # Using slots for reduced memory usage and improved performance.
    _tag_dataclass_kwargs = dict(frozen=True, repr=False, kw_only=True, slots=True)
else:
    # deprecated: description='always use dataclass slots and keyword-only args' python_version='3.9'
    _tag_dataclass_kwargs = dict(frozen=True, repr=False)

_T = t.TypeVar('_T')
_TAnsibleSerializable = t.TypeVar('_TAnsibleSerializable', bound='AnsibleSerializable')
_TAnsibleDatatagBase = t.TypeVar('_TAnsibleDatatagBase', bound='AnsibleDatatagBase')
_TAnsibleTaggedObject = t.TypeVar('_TAnsibleTaggedObject', bound='AnsibleTaggedObject')

_NO_INSTANCE_STORAGE = t.cast(t.Tuple[str], tuple())
_ANSIBLE_TAGGED_OBJECT_SLOTS = tuple(('_ansible_tags_mapping',))

# shared empty frozenset for default values
_empty_frozenset: t.FrozenSet = frozenset()

# Technical Notes
#
# Tagged values compare (and thus hash) the same as their base types, so a value that differs only by its tags will appear identical to non-tag-aware code.
# This will affect storage and update of tagged values in dictionary keys, sets, etc. While tagged values can be used as keys in hashable collections,
# updating a key usually requires removal and re-addition.


class AnsibleTagHelper:
    """Utility methods for working with Ansible data tags."""

    @staticmethod
    def untag(value: _T, *tag_types: t.Type[AnsibleDatatagBase]) -> _T:
        """
        If tags matching any of `tag_types` are present on `value`, return a copy with those tags removed.
        If no `tag_types` are specified and the object has tags, return a copy with all tags removed.
        Otherwise, the original `value` is returned.
        """
        tag_set = AnsibleTagHelper.tags(value)

        if not tag_set:
            return value

        if tag_types:
            tags_mapping = _AnsibleTagsMapping((type(tag), tag) for tag in tag_set if type(tag) not in tag_types)  # pylint: disable=unidiomatic-typecheck

            if len(tags_mapping) == len(tag_set):
                return value  # if no tags were removed, return the original instance
        else:
            tags_mapping = None

        if not tags_mapping:
            if t.cast(AnsibleTaggedObject, value)._empty_tags_as_native:
                return t.cast(AnsibleTaggedObject, value)._native_copy()

            tags_mapping = _EMPTY_INTERNAL_TAGS_MAPPING

        tagged_type = AnsibleTaggedObject._get_tagged_type(type(value))

        return t.cast(_T, tagged_type._instance_factory(value, tags_mapping))

    @staticmethod
    def tags(value: t.Any) -> t.FrozenSet[AnsibleDatatagBase]:
        tags = _try_get_internal_tags_mapping(value)

        if tags is _EMPTY_INTERNAL_TAGS_MAPPING:
            return _empty_frozenset

        return frozenset(tags.values())

    @staticmethod
    def tag_types(value: t.Any) -> t.FrozenSet[t.Type[AnsibleDatatagBase]]:
        tags = _try_get_internal_tags_mapping(value)

        if tags is _EMPTY_INTERNAL_TAGS_MAPPING:
            return _empty_frozenset

        return frozenset(tags)

    @staticmethod
    def base_type(type_or_value: t.Any, /) -> type:
        """Return the friendly type of the given type or value. If the type is an AnsibleTaggedObject, the native type will be used."""
        if isinstance(type_or_value, type):
            the_type = type_or_value
        else:
            the_type = type(type_or_value)

        if issubclass(the_type, AnsibleTaggedObject):
            the_type = type_or_value._native_type

        # DTFIX-FUTURE: provide a knob to optionally report the real type for debugging purposes
        return the_type

    @staticmethod
    def as_native_type(value: _T) -> _T:
        """
        Returns an untagged native data type matching the input value, or the original input if the value was not a tagged type.
        Containers are not recursively processed.
        """
        if isinstance(value, AnsibleTaggedObject):
            value = value._native_copy()

        return value

    @staticmethod
    @t.overload
    def tag_copy(src: t.Any, value: _T) -> _T: ...  # pragma: nocover

    @staticmethod
    @t.overload
    def tag_copy(src: t.Any, value: t.Any, *, value_type: type[_T]) -> _T: ...  # pragma: nocover

    @staticmethod
    @t.overload
    def tag_copy(src: t.Any, value: _T, *, value_type: None = None) -> _T: ...  # pragma: nocover

    @staticmethod
    def tag_copy(src: t.Any, value: _T, *, value_type: t.Optional[type] = None) -> _T:
        """Return a copy of `value`, with tags copied from `src`, overwriting any existing tags of the same types."""
        src_tags = AnsibleTagHelper.tags(src)
        value_tags = [(tag, tag._get_tag_to_propagate(src, value, value_type=value_type)) for tag in src_tags]
        tags = [tag[1] for tag in value_tags if tag[1] is not None]
        tag_types_to_remove = [type(tag[0]) for tag in value_tags if tag[1] is None]

        if tag_types_to_remove:
            value = AnsibleTagHelper.untag(value, *tag_types_to_remove)

        return AnsibleTagHelper.tag(value, tags, value_type=value_type)

    @staticmethod
    @t.overload
    def tag(value: _T, tags: t.Union[AnsibleDatatagBase, t.Iterable[AnsibleDatatagBase]]) -> _T: ...  # pragma: nocover

    @staticmethod
    @t.overload
    def tag(value: t.Any, tags: t.Union[AnsibleDatatagBase, t.Iterable[AnsibleDatatagBase]], *, value_type: type[_T]) -> _T: ...  # pragma: nocover

    @staticmethod
    @t.overload
    def tag(value: _T, tags: t.Union[AnsibleDatatagBase, t.Iterable[AnsibleDatatagBase]], *, value_type: None = None) -> _T: ...  # pragma: nocover

    @staticmethod
    def tag(value: _T, tags: t.Union[AnsibleDatatagBase, t.Iterable[AnsibleDatatagBase]], *, value_type: t.Optional[type] = None) -> _T:
        """
        Return a copy of `value`, with `tags` applied, overwriting any existing tags of the same types.
        If `value` is an ignored type, or `tags` is empty, the original `value` will be returned.
        If `value` is not taggable, a `NotTaggableError` exception will be raised.
        If `value_type` was given, that type will be returned instead.
        """
        if value_type is None:
            value_type_specified = False
            value_type = type(value)
        else:
            value_type_specified = True

        # if no tags to apply, just return what we got
        # NB: this only works because the untaggable types are singletons (and thus direct type comparison works)
        if not tags or value_type in _untaggable_types:
            if value_type_specified:
                return value_type(value)

            return value

        tag_list: list[AnsibleDatatagBase]

        # noinspection PyProtectedMember
        if type(tags) in _known_tag_types:
            tag_list = [tags]  # type: ignore[list-item]
        else:
            tag_list = list(tags)  # type: ignore[arg-type]

            for idx, tag in enumerate(tag_list):
                # noinspection PyProtectedMember
                if type(tag) not in _known_tag_types:
                    # noinspection PyProtectedMember
                    raise TypeError(f'tags[{idx}] of type {type(tag)} is not one of {_known_tag_types}')

        existing_internal_tags_mapping = _try_get_internal_tags_mapping(value)

        if existing_internal_tags_mapping is not _EMPTY_INTERNAL_TAGS_MAPPING:
            # include the existing tags first so new tags of the same type will overwrite
            tag_list = list(chain(existing_internal_tags_mapping.values(), tag_list))

        tags_mapping = _AnsibleTagsMapping((type(tag), tag) for tag in tag_list)
        tagged_type = AnsibleTaggedObject._get_tagged_type(value_type)

        return t.cast(_T, tagged_type._instance_factory(value, tags_mapping))

    @staticmethod
    def try_tag(value: _T, tags: t.Union[AnsibleDatatagBase, t.Iterable[AnsibleDatatagBase]]) -> _T:
        """
        Return a copy of `value`, with `tags` applied, overwriting any existing tags of the same types.
        If `value` is not taggable or `tags` is empty, the original `value` will be returned.
        """
        try:
            return AnsibleTagHelper.tag(value, tags)
        except NotTaggableError:
            return value


class AnsibleSerializable:
    __slots__ = _NO_INSTANCE_STORAGE

    _known_type_map: t.ClassVar[t.Dict[str, t.Type['AnsibleSerializable']]] = {}
    _TYPE_KEY: t.ClassVar[str] = '__ansible_type'

    _type_key: t.ClassVar[str]

    def __init_subclass__(cls, **kwargs) -> None:
        # this is needed to call __init__subclass__ on mixins for derived types
        super().__init_subclass__(**kwargs)

        cls._type_key = cls.__name__

        # DTFIX-FUTURE: is there a better way to exclude non-abstract types which are base classes?
        if not inspect.isabstract(cls) and not cls.__name__.endswith('Base') and cls.__name__ != 'AnsibleTaggedObject':
            AnsibleSerializable._known_type_map[cls._type_key] = cls

    @classmethod
    @abc.abstractmethod
    def _from_dict(cls: t.Type[_TAnsibleSerializable], d: t.Dict[str, t.Any]) -> object:
        """Return an instance of this type, created from the given dictionary."""

    @abc.abstractmethod
    def _as_dict(self) -> t.Dict[str, t.Any]:
        """
        Return a serialized version of this instance as a dictionary.
        This operation is *NOT* recursive - the returned dictionary may still include custom types.
        It is the responsibility of the caller to handle recursion of the returned dict.
        """

    def _serialize(self) -> t.Dict[str, t.Any]:
        value = self._as_dict()
        value.update({AnsibleSerializable._TYPE_KEY: self._type_key})

        return value

    @staticmethod
    def _deserialize(data: t.Dict[str, t.Any]) -> object:
        """Deserialize an object from the supplied data dict, which will be mutated if it contains a type key."""
        type_name = data.pop(AnsibleSerializable._TYPE_KEY, ...)  # common usage assumes `data` is an intermediate dict provided by a deserializer

        if type_name is ...:
            return None

        type_value = AnsibleSerializable._known_type_map.get(type_name)

        if not type_value:
            raise ValueError(f'An unknown {AnsibleSerializable._TYPE_KEY!r} value {type_name!r} was encountered during deserialization.')

        return type_value._from_dict(data)

    def _repr(self, name: str) -> str:
        args = self._as_dict()
        arg_string = ', '.join((f'{k}={v!r}' for k, v in args.items()))
        return f'{name}({arg_string})'


class AnsibleSerializableEnum(AnsibleSerializable, enum.Enum):
    """Base class for serializable enumerations."""

    def _as_dict(self) -> t.Dict[str, t.Any]:
        return dict(value=self.value)

    @classmethod
    def _from_dict(cls, d: t.Dict[str, t.Any]) -> t.Self:
        return cls(d['value'].lower())

    def __str__(self) -> str:
        return self.value

    def __repr__(self) -> str:
        return f'<{self.__class__.__name__}.{self.name}>'

    @staticmethod
    def _generate_next_value_(name, start, count, last_values):
        return name.lower()


class AnsibleSerializableWrapper(AnsibleSerializable, t.Generic[_T], metaclass=abc.ABCMeta):
    __slots__ = ('_value',)

    _wrapped_types: t.ClassVar[dict[type, type[AnsibleSerializable]]] = {}
    _wrapped_type: t.ClassVar[type] = type(None)

    def __init__(self, value: _T) -> None:
        self._value: _T = value

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)

        cls._wrapped_type = t.get_args(cls.__orig_bases__[0])[0]
        cls._wrapped_types[cls._wrapped_type] = cls


class AnsibleSerializableDate(AnsibleSerializableWrapper[datetime.date]):
    __slots__ = _NO_INSTANCE_STORAGE

    @classmethod
    def _from_dict(cls: t.Type[_TAnsibleSerializable], d: t.Dict[str, t.Any]) -> datetime.date:
        return datetime.date.fromisoformat(d['iso8601'])

    def _as_dict(self) -> t.Dict[str, t.Any]:
        return dict(
            iso8601=self._value.isoformat(),
        )


class AnsibleSerializableTime(AnsibleSerializableWrapper[datetime.time]):
    __slots__ = _NO_INSTANCE_STORAGE

    @classmethod
    def _from_dict(cls: t.Type[_TAnsibleSerializable], d: t.Dict[str, t.Any]) -> datetime.time:
        value = datetime.time.fromisoformat(d['iso8601'])
        value.replace(fold=d['fold'])

        return value

    def _as_dict(self) -> t.Dict[str, t.Any]:
        return dict(
            iso8601=self._value.isoformat(),
            fold=self._value.fold,
        )


class AnsibleSerializableDateTime(AnsibleSerializableWrapper[datetime.datetime]):
    __slots__ = _NO_INSTANCE_STORAGE

    @classmethod
    def _from_dict(cls: t.Type[_TAnsibleSerializable], d: t.Dict[str, t.Any]) -> datetime.datetime:
        value = datetime.datetime.fromisoformat(d['iso8601'])
        value.replace(fold=d['fold'])

        return value

    def _as_dict(self) -> t.Dict[str, t.Any]:
        return dict(
            iso8601=self._value.isoformat(),
            fold=self._value.fold,
        )


@dataclasses.dataclass(**_tag_dataclass_kwargs)
class AnsibleSerializableDataclass(AnsibleSerializable, metaclass=abc.ABCMeta):
    _validation_allow_subclasses = True
    _validation_auto_enabled = True

    def _as_dict(self) -> t.Dict[str, t.Any]:
        # omit None values when None is the field default
        # DTFIX-FUTURE: this implementation means we can never change the default on fields which have None for their default
        #          other defaults can be changed -- but there's no way to override this behavior either way for other default types
        #          it's a trip hazard to have the default logic here, rather than per field (or not at all)
        #          consider either removing the filtering or requiring it to be explicitly set per field using dataclass metadata
        fields = ((field, getattr(self, field.name)) for field in dataclasses.fields(self))
        return {field.name: value for field, value in fields if value is not None or field.default is not None}

    @classmethod
    def _from_dict(cls, d: t.Dict[str, t.Any]) -> t.Self:
        # DTFIX-FUTURE: optimize this to avoid the dataclasses fields metadata and get_origin stuff at runtime
        type_hints = t.get_type_hints(cls)
        mutated_dict: dict[str, t.Any] | None = None

        for field in dataclasses.fields(cls):
            if t.get_origin(type_hints[field.name]) is tuple:  # NOTE: only supports bare tuples, not optional or inside a union
                if type(field_value := d.get(field.name)) is list:  # pylint: disable=unidiomatic-typecheck
                    if mutated_dict is None:
                        mutated_dict = d.copy()

                    mutated_dict[field.name] = tuple(field_value)

        return cls(**(mutated_dict or d))

    def __init_subclass__(cls, **kwargs) -> None:
        super(AnsibleSerializableDataclass, cls).__init_subclass__(**kwargs)  # cannot use super() without arguments when using slots

        if cls._validation_auto_enabled:
            try:
                _dataclass_validation.inject_post_init_validation(cls, cls._validation_allow_subclasses)  # code gen a real __post_init__ method
            except Exception as ex:
                raise Exception(f'Validation code generation failed on {cls}.') from ex


class Tripwire:
    """Marker mixin for types that should raise an error when encountered."""

    __slots__ = _NO_INSTANCE_STORAGE

    def trip(self) -> t.NoReturn:
        """Derived types should implement a failure behavior."""
        raise NotImplementedError()


@dataclasses.dataclass(**_tag_dataclass_kwargs)
class AnsibleDatatagBase(AnsibleSerializableDataclass, metaclass=abc.ABCMeta):
    """
    Base class for data tagging tag types.
    New tag types need to be considered very carefully; e.g.: which serialization/runtime contexts they're allowed in, fallback behavior, propagation.
    """

    _validation_allow_subclasses = False

    def __init_subclass__(cls, **kwargs) -> None:
        # NOTE: This method is called twice when the datatag type is a dataclass.
        super(AnsibleDatatagBase, cls).__init_subclass__(**kwargs)  # cannot use super() without arguments when using slots

        # DTFIX-FUTURE: "freeze" this after module init has completed to discourage custom external tag subclasses

        # DTFIX-FUTURE: is there a better way to exclude non-abstract types which are base classes?
        if not inspect.isabstract(cls) and not cls.__name__.endswith('Base'):
            existing = _known_tag_type_map.get(cls.__name__)

            if existing:
                # When the datatag type is a dataclass, the first instance will be the non-dataclass type.
                # It must be removed from the known tag types before adding the dataclass version.
                _known_tag_types.remove(existing)

            _known_tag_type_map[cls.__name__] = cls
            _known_tag_types.add(cls)

    @classmethod
    def is_tagged_on(cls, value: t.Any) -> bool:
        return cls in _try_get_internal_tags_mapping(value)

    @classmethod
    def first_tagged_on(cls, *values: t.Any) -> t.Any | None:
        """Return the first value which is tagged with this type, or None if no match is found."""
        for value in values:
            if cls.is_tagged_on(value):
                return value

        return None

    @classmethod
    def get_tag(cls, value: t.Any) -> t.Optional[t.Self]:
        return _try_get_internal_tags_mapping(value).get(cls)

    @classmethod
    def get_required_tag(cls, value: t.Any) -> t.Self:
        if (tag := cls.get_tag(value)) is None:
            # DTFIX-FUTURE: we really should have a way to use AnsibleError with obj in module_utils when it's controller-side
            raise ValueError(f'The type {type(value).__name__!r} is not tagged with {cls.__name__!r}.')

        return tag

    @classmethod
    def untag(cls, value: _T) -> _T:
        """
        If this tag type is present on `value`, return a copy with that tag removed.
        Otherwise, the original `value` is returned.
        """
        return AnsibleTagHelper.untag(value, cls)

    def tag(self, value: _T) -> _T:
        """
        Return a copy of `value` with this tag applied, overwriting any existing tag of the same type.
        If `value` is an ignored type, the original `value` will be returned.
        If `value` is not taggable, a `NotTaggableError` exception will be raised.
        """
        return AnsibleTagHelper.tag(value, self)

    def try_tag(self, value: _T) -> _T:
        """
        Return a copy of `value` with this tag applied, overwriting any existing tag of the same type.
        If `value` is not taggable, the original `value` will be returned.
        """
        return AnsibleTagHelper.try_tag(value, self)

    def _get_tag_to_propagate(self, src: t.Any, value: object, *, value_type: t.Optional[type] = None) -> t.Self | None:
        """
        Called by `AnsibleTagHelper.tag_copy` during tag propagation.
        Returns an instance of this tag appropriate for propagation to `value`, or `None` if the tag should not be propagated.
        Derived implementations may consult the arguments relayed from `tag_copy` to determine if and how the tag should be propagated.
        """
        return self

    def __repr__(self) -> str:
        return AnsibleSerializable._repr(self, self.__class__.__name__)


# used by the datatag Ansible/Jinja test plugin to find tags by name
_known_tag_type_map: t.Dict[str, t.Type[AnsibleDatatagBase]] = {}
_known_tag_types: t.Set[t.Type[AnsibleDatatagBase]] = set()

if sys.version_info >= (3, 9):
    # Include the key and value types in the type hints on Python 3.9 and later.
    # Earlier versions do not support subscriptable dict.
    # deprecated: description='always use subscriptable dict' python_version='3.8'
    class _AnsibleTagsMapping(dict[type[_TAnsibleDatatagBase], _TAnsibleDatatagBase]):
        __slots__ = _NO_INSTANCE_STORAGE

else:

    class _AnsibleTagsMapping(dict):
        __slots__ = _NO_INSTANCE_STORAGE


class _EmptyROInternalTagsMapping(dict):
    """
    Optimizes empty tag mapping by using a shared singleton read-only dict.
    Since mappingproxy is not pickle-able and causes other problems, we had to roll our own.
    """

    def __new__(cls):
        try:
            # noinspection PyUnresolvedReferences
            return cls._instance
        except AttributeError:
            cls._instance = dict.__new__(cls)

        # noinspection PyUnresolvedReferences
        return cls._instance

    def __setitem__(self, key, value):
        raise NotImplementedError()

    def setdefault(self, __key, __default=None):
        raise NotImplementedError()

    def update(self, __m, **kwargs):
        raise NotImplementedError()


_EMPTY_INTERNAL_TAGS_MAPPING = t.cast(_AnsibleTagsMapping, _EmptyROInternalTagsMapping())
"""
An empty read-only mapping of tags.
Also used as a sentinel to cheaply determine that a type is not tagged by using a reference equality check.
"""


class CollectionWithMro(c.Collection, t.Protocol):
    """Used to represent a Collection with __mro__ in a TypeGuard for tools that don't include __mro__ in Collection."""

    __mro__: tuple[type, ...]


def is_non_scalar_collection_type(value: type) -> t.TypeGuard[type[CollectionWithMro]]:
    """Returns True if the value is a non-scalar collection type, otherwise returns False."""
    return issubclass(value, c.Collection) and not issubclass(value, str) and not issubclass(value, bytes)


def _try_get_internal_tags_mapping(value: t.Any) -> _AnsibleTagsMapping:
    """Return the internal tag mapping of the given value, or a sentinel value if it is not tagged."""
    # noinspection PyBroadException
    try:
        # noinspection PyProtectedMember
        tags = value._ansible_tags_mapping
    except Exception:
        # try/except is a cheap way to determine if this is a tagged object without using isinstance
        # handling Exception accounts for types that may raise something other than AttributeError
        return _EMPTY_INTERNAL_TAGS_MAPPING

    # handle cases where the instance always returns something, such as Marker or MagicMock
    if type(tags) is not _AnsibleTagsMapping:  # pylint: disable=unidiomatic-typecheck
        return _EMPTY_INTERNAL_TAGS_MAPPING

    return tags


class NotTaggableError(TypeError):
    def __init__(self, value):
        super(NotTaggableError, self).__init__('{} is not taggable'.format(value))


@dataclasses.dataclass(**_tag_dataclass_kwargs)
class AnsibleSingletonTagBase(AnsibleDatatagBase):
    def __new__(cls):
        try:
            # noinspection PyUnresolvedReferences
            return cls._instance
        except AttributeError:
            cls._instance = AnsibleDatatagBase.__new__(cls)

        # noinspection PyUnresolvedReferences
        return cls._instance

    def _as_dict(self) -> t.Dict[str, t.Any]:
        return {}


class AnsibleTaggedObject(AnsibleSerializable):
    __slots__ = _NO_INSTANCE_STORAGE

    _native_type: t.ClassVar[type]
    _item_source: t.ClassVar[t.Optional[t.Callable]] = None

    _tagged_type_map: t.ClassVar[t.Dict[type, t.Type['AnsibleTaggedObject']]] = {}
    _tagged_collection_types: t.ClassVar[t.Set[t.Type[c.Collection]]] = set()
    _collection_types: t.ClassVar[t.Set[t.Type[c.Collection]]] = set()

    _empty_tags_as_native: t.ClassVar[bool] = True  # by default, untag will revert to the native type when no tags remain
    _subclasses_native_type: t.ClassVar[bool] = True  # by default, tagged types are assumed to subclass the type they augment

    _ansible_tags_mapping: _AnsibleTagsMapping | _EmptyROInternalTagsMapping = _EMPTY_INTERNAL_TAGS_MAPPING
    """
    Efficient internal storage of tags, indexed by tag type.
    Contains no more than one instance of each tag type.
    This is defined as a class attribute to support type hinting and documentation.
    It is overwritten with an instance attribute during instance creation.
    The instance attribute slot is provided by the derived type.
    """

    def __init_subclass__(cls, **kwargs) -> None:
        super().__init_subclass__(**kwargs)

        try:
            init_class = cls._init_class  # type: ignore[attr-defined]
        except AttributeError:
            pass
        else:
            init_class()

        if not cls._subclasses_native_type:
            return  # NOTE: When not subclassing a native type, the derived type must set cls._native_type itself and cls._empty_tags_as_native to False.

        try:
            # Subclasses of tagged types will already have a native type set and won't need to detect it.
            # Special types which do not subclass a native type can also have their native type already set.
            # Automatic item source selection is only implemented for types that don't set _native_type.
            cls._native_type
        except AttributeError:
            # Direct subclasses of native types won't have cls._native_type set, so detect the native type.
            cls._native_type = cls.__bases__[0]

            # Detect the item source if not already set.
            if cls._item_source is None and is_non_scalar_collection_type(cls._native_type):
                cls._item_source = cls._native_type.__iter__  # type: ignore[attr-defined]

        # Use a collection specific factory for types with item sources.
        if cls._item_source:
            cls._instance_factory = cls._instance_factory_collection  # type: ignore[method-assign]

        new_type_direct_subclass = cls.__mro__[1]

        conflicting_impl = AnsibleTaggedObject._tagged_type_map.get(new_type_direct_subclass)

        if conflicting_impl:
            raise TypeError(f'Cannot define type {cls.__name__!r} since {conflicting_impl.__name__!r} already extends {new_type_direct_subclass.__name__!r}.')

        AnsibleTaggedObject._tagged_type_map[new_type_direct_subclass] = cls

        if is_non_scalar_collection_type(cls):
            AnsibleTaggedObject._tagged_collection_types.add(cls)
            AnsibleTaggedObject._collection_types.update({cls, new_type_direct_subclass})

    def _native_copy(self) -> t.Any:
        """
        Returns a copy of the current instance as its native Python type.
        Any dynamic access behaviors that apply to this instance will be used during creation of the copy.
        In the case of a container type, this is a shallow copy.
        Recursive calls to native_copy are the responsibility of the caller.
        """
        return self._native_type(self)  # pylint: disable=abstract-class-instantiated

    @classmethod
    def _instance_factory(cls, value: t.Any, tags_mapping: _AnsibleTagsMapping) -> t.Self:
        # There's no way to indicate cls is callable with a single arg without defining a useless __init__.
        instance = cls(value)  # type: ignore[call-arg]
        instance._ansible_tags_mapping = tags_mapping

        return instance

    @staticmethod
    def _get_tagged_type(value_type: type) -> type[AnsibleTaggedObject]:
        tagged_type: t.Optional[type[AnsibleTaggedObject]]

        if issubclass(value_type, AnsibleTaggedObject):
            tagged_type = value_type
        else:
            tagged_type = AnsibleTaggedObject._tagged_type_map.get(value_type)

        if not tagged_type:
            raise NotTaggableError(value_type)

        return tagged_type

    def _as_dict(self) -> t.Dict[str, t.Any]:
        return dict(
            value=self._native_copy(),
            tags=list(self._ansible_tags_mapping.values()),
        )

    @classmethod
    def _from_dict(cls: t.Type[_TAnsibleTaggedObject], d: t.Dict[str, t.Any]) -> _TAnsibleTaggedObject:
        return AnsibleTagHelper.tag(**d)

    @classmethod
    def _instance_factory_collection(
        cls,
        value: t.Any,
        tags_mapping: _AnsibleTagsMapping,
    ) -> t.Self:
        if type(value) in AnsibleTaggedObject._collection_types:
            # use the underlying iterator to avoid access/iteration side effects (e.g. templating/wrapping on Lazy subclasses)
            instance = cls(cls._item_source(value))  # type: ignore[call-arg,misc]
        else:
            # this is used when the value is a generator
            instance = cls(value)  # type: ignore[call-arg]

        instance._ansible_tags_mapping = tags_mapping

        return instance

    def _copy_collection(self) -> AnsibleTaggedObject:
        """
        Return a shallow copy of this instance, which must be a collection.
        This uses the underlying iterator to avoid access/iteration side effects (e.g. templating/wrapping on Lazy subclasses).
        """
        return AnsibleTagHelper.tag_copy(self, type(self)._item_source(self), value_type=type(self))  # type: ignore[misc]

    @classmethod
    def _new(cls, value: t.Any, *args, **kwargs) -> t.Self:
        if type(value) is _AnsibleTagsMapping:  # pylint: disable=unidiomatic-typecheck
            self = cls._native_type.__new__(cls, *args, **kwargs)
            self._ansible_tags_mapping = value
            return self

        return cls._native_type.__new__(cls, value, *args, **kwargs)

    def _reduce(self, reduced: t.Union[str, tuple[t.Any, ...]]) -> tuple:
        if type(reduced) is not tuple:  # pylint: disable=unidiomatic-typecheck
            raise TypeError()

        updated: list[t.Any] = list(reduced)
        updated[1] = (self._ansible_tags_mapping,) + updated[1]

        return tuple(updated)


class _AnsibleTaggedStr(str, AnsibleTaggedObject):
    __slots__ = _ANSIBLE_TAGGED_OBJECT_SLOTS


class _AnsibleTaggedBytes(bytes, AnsibleTaggedObject):
    # nonempty __slots__ not supported for subtype of 'bytes'
    pass


class _AnsibleTaggedInt(int, AnsibleTaggedObject):
    # nonempty __slots__ not supported for subtype of 'int'
    pass


class _AnsibleTaggedFloat(float, AnsibleTaggedObject):
    __slots__ = _ANSIBLE_TAGGED_OBJECT_SLOTS


class _AnsibleTaggedDateTime(datetime.datetime, AnsibleTaggedObject):
    __slots__ = _ANSIBLE_TAGGED_OBJECT_SLOTS

    @classmethod
    def _instance_factory(cls, value: datetime.datetime, tags_mapping: _AnsibleTagsMapping) -> _AnsibleTaggedDateTime:
        instance = cls(
            year=value.year,
            month=value.month,
            day=value.day,
            hour=value.hour,
            minute=value.minute,
            second=value.second,
            microsecond=value.microsecond,
            tzinfo=value.tzinfo,
            fold=value.fold,
        )

        instance._ansible_tags_mapping = tags_mapping

        return instance

    def _native_copy(self) -> datetime.datetime:
        return datetime.datetime(
            year=self.year,
            month=self.month,
            day=self.day,
            hour=self.hour,
            minute=self.minute,
            second=self.second,
            microsecond=self.microsecond,
            tzinfo=self.tzinfo,
            fold=self.fold,
        )

    def __new__(cls, year, *args, **kwargs):
        return super()._new(year, *args, **kwargs)

    def __reduce_ex__(self, protocol: t.SupportsIndex) -> tuple:
        return super()._reduce(super().__reduce_ex__(protocol))

    def __repr__(self) -> str:
        return self._native_copy().__repr__()


class _AnsibleTaggedDate(datetime.date, AnsibleTaggedObject):
    __slots__ = _ANSIBLE_TAGGED_OBJECT_SLOTS

    @classmethod
    def _instance_factory(cls, value: datetime.date, tags_mapping: _AnsibleTagsMapping) -> _AnsibleTaggedDate:
        instance = cls(
            year=value.year,
            month=value.month,
            day=value.day,
        )

        instance._ansible_tags_mapping = tags_mapping

        return instance

    def _native_copy(self) -> datetime.date:
        return datetime.date(
            year=self.year,
            month=self.month,
            day=self.day,
        )

    def __new__(cls, year, *args, **kwargs):
        return super()._new(year, *args, **kwargs)

    def __reduce__(self) -> tuple:
        return super()._reduce(super().__reduce__())

    def __repr__(self) -> str:
        return self._native_copy().__repr__()


class _AnsibleTaggedTime(datetime.time, AnsibleTaggedObject):
    __slots__ = _ANSIBLE_TAGGED_OBJECT_SLOTS

    @classmethod
    def _instance_factory(cls, value: datetime.time, tags_mapping: _AnsibleTagsMapping) -> _AnsibleTaggedTime:
        instance = cls(
            hour=value.hour,
            minute=value.minute,
            second=value.second,
            microsecond=value.microsecond,
            tzinfo=value.tzinfo,
            fold=value.fold,
        )

        instance._ansible_tags_mapping = tags_mapping

        return instance

    def _native_copy(self) -> datetime.time:
        return datetime.time(
            hour=self.hour,
            minute=self.minute,
            second=self.second,
            microsecond=self.microsecond,
            tzinfo=self.tzinfo,
            fold=self.fold,
        )

    def __new__(cls, hour, *args, **kwargs):
        return super()._new(hour, *args, **kwargs)

    def __reduce_ex__(self, protocol: t.SupportsIndex) -> tuple:
        return super()._reduce(super().__reduce_ex__(protocol))

    def __repr__(self) -> str:
        return self._native_copy().__repr__()


class _AnsibleTaggedDict(dict, AnsibleTaggedObject):
    __slots__ = _ANSIBLE_TAGGED_OBJECT_SLOTS

    _item_source: t.ClassVar[t.Optional[t.Callable]] = dict.items

    def __copy__(self):
        return super()._copy_collection()

    def copy(self) -> _AnsibleTaggedDict:
        return copy.copy(self)

    # NB: Tags are intentionally not preserved for operator methods that return a new instance. In-place operators ignore tags from the `other` instance.
    # Propagation of tags in these cases is left to the caller, based on needs specific to their use case.


class _AnsibleTaggedList(list, AnsibleTaggedObject):
    __slots__ = _ANSIBLE_TAGGED_OBJECT_SLOTS

    def __copy__(self):
        return super()._copy_collection()

    def copy(self) -> _AnsibleTaggedList:
        return copy.copy(self)

    # NB: Tags are intentionally not preserved for operator methods that return a new instance. In-place operators ignore tags from the `other` instance.
    # Propagation of tags in these cases is left to the caller, based on needs specific to their use case.


class _AnsibleTaggedSet(set, AnsibleTaggedObject):
    __slots__ = _ANSIBLE_TAGGED_OBJECT_SLOTS

    def __copy__(self):
        return super()._copy_collection()

    def copy(self):
        return copy.copy(self)

    def __init__(self, value=None, *args, **kwargs):
        if type(value) is _AnsibleTagsMapping:  # pylint: disable=unidiomatic-typecheck
            super().__init__(*args, **kwargs)
        else:
            super().__init__(value, *args, **kwargs)

    def __new__(cls, value=None, *args, **kwargs):
        return super()._new(value, *args, **kwargs)

    def __reduce_ex__(self, protocol: t.SupportsIndex) -> tuple:
        return super()._reduce(super().__reduce_ex__(protocol))

    def __str__(self) -> str:
        return self._native_copy().__str__()

    def __repr__(self) -> str:
        return self._native_copy().__repr__()


class _AnsibleTaggedTuple(tuple, AnsibleTaggedObject):
    # nonempty __slots__ not supported for subtype of 'tuple'

    def __copy__(self):
        return super()._copy_collection()


_untaggable_types = {type(None), bool}
"""
Attempts to apply tags to values of these types will be silently ignored.
While we could proxy or subclass builtin singletons, they're idiomatically compared with "is" reference equality, which we can't customize.
This set gets augmented with additional types when some controller-only types are imported.
"""

# noinspection PyProtectedMember
_ANSIBLE_ALLOWED_VAR_TYPES = frozenset({type(None), bool}) | set(AnsibleTaggedObject._tagged_type_map) | set(AnsibleTaggedObject._tagged_type_map.values())
"""These are the exact types supported by Ansible's variable storage."""

_ANSIBLE_ALLOWED_NON_SCALAR_COLLECTION_VAR_TYPES = frozenset(item for item in _ANSIBLE_ALLOWED_VAR_TYPES if is_non_scalar_collection_type(item))
"""These are the exact non-scalar collection types supported by Ansible's variable storage."""

_ANSIBLE_ALLOWED_MAPPING_VAR_TYPES = frozenset(item for item in _ANSIBLE_ALLOWED_VAR_TYPES if issubclass(item, c.Mapping))
"""These are the exact mapping types supported by Ansible's variable storage."""

_ANSIBLE_ALLOWED_SCALAR_VAR_TYPES = _ANSIBLE_ALLOWED_VAR_TYPES - _ANSIBLE_ALLOWED_NON_SCALAR_COLLECTION_VAR_TYPES
"""These are the exact scalar types supported by Ansible's variable storage."""



#!/usr/bin/python
from __future__ import annotations

from ansible.module_utils.basic import AnsibleModule


def main():
    module = AnsibleModule(argument_spec=dict(
        secret=dict(no_log=True, type='raw'),
        notsecret=dict(no_log=False, type='raw'),
    ))

    msg = "My secret is: (%s), but don't tell %s" % (module.params['secret'], module.params['notsecret'])
    module.exit_json(
        msg=msg,
        changed=bool(module.params['secret'] == module.params['notsecret']),
        secret=module.params['secret'],
        notsecret=module.params['notsecret'],
    )


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
# Copyright (c) 2019 Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import annotations

from ansible.module_utils.facts.network import fc_wwn
import pytest


# AIX lsdev
LSDEV_OUTPUT = """
fcs0 Defined   00-00 8Gb PCI Express Dual Port FC Adapter (df1000f114108a03)
fcs1 Available 04-00 8Gb PCI Express Dual Port FC Adapter (df1000f114108a03)
"""

# slightly cut output of lscfg (from Z0 to ZC)
LSCFG_OUTPUT = """
  fcs1             U78CB.001.WZS00ZS-P1-C9-T1  8Gb PCI Express Dual Port FC Adapter (df1000f114108a03)

        Part Number.................00E0806
        Serial Number...............1C4090830F
        Manufacturer................001C
        EC Level.................... D77161
        Customer Card ID Number.....577D
        FRU Number..................00E0806
        Device Specific.(ZM)........3
        Network Address.............10000090FA551508
        ROS Level and ID............027820B7
        Device Specific.(Z0)........31004549
        Device Specific.(ZC)........00000000
        Hardware Location Code......U78CB.001.WZS00ZS-P1-C9-T1
"""

# Solaris
FCINFO_OUTPUT = """
HBA Port WWN: 10000090fa1658de
        Port Mode: Initiator
        Port ID: 30100
        OS Device Name: /dev/cfg/c13
        Manufacturer: Emulex
        Model: LPe12002-S
        Firmware Version: LPe12002-S 2.01a12
        FCode/BIOS Version: Boot:5.03a0 Fcode:3.01a1
        Serial Number: 4925381+13090001ER
        Driver Name: emlxs
        Driver Version: 3.3.00.1 (2018.01.05.16.30)
        Type: N-port
        State: online
        Supported Speeds: 2Gb 4Gb 8Gb
        Current Speed: 8Gb
        Node WWN: 20000090fa1658de
        NPIV Not Supported
"""

IOSCAN_OUT = """
Class     I  H/W Path    Driver S/W State   H/W Type     Description
==================================================================
fc        0  2/0/10/1/0  fcd   CLAIMED     INTERFACE    HP AB379-60101 4Gb Dual Port PCI/PCI-X Fibre Channel Adapter (FC Port 1)
                        /dev/fcd0
"""

FCMSUTIL_OUT = """
                           Vendor ID is = 0x1077
                           Device ID is = 0x2422
            PCI Sub-system Vendor ID is = 0x103C
                   PCI Sub-system ID is = 0x12D7
                               PCI Mode = PCI-X 133 MHz
                       ISP Code version = 5.4.0
                       ISP Chip version = 3
                               Topology = PTTOPT_FABRIC
                             Link Speed = 4Gb
                     Local N_Port_id is = 0x010300
                  Previous N_Port_id is = None
            N_Port Node World Wide Name = 0x50060b00006975ed
            N_Port Port World Wide Name = 0x50060b00006975ec
            Switch Port World Wide Name = 0x200300051e046c0f
            Switch Node World Wide Name = 0x100000051e046c0f
              N_Port Symbolic Port Name = server1_fcd0
              N_Port Symbolic Node Name = server1_HP-UX_B.11.31
                           Driver state = ONLINE
                       Hardware Path is = 2/0/10/1/0
                     Maximum Frame Size = 2048
         Driver-Firmware Dump Available = NO
         Driver-Firmware Dump Timestamp = N/A
                                   TYPE = PFC
                         NPIV Supported = YES
                         Driver Version = @(#) fcd B.11.31.1103 Dec  6 2010
"""


def mock_get_bin_path(cmd, required=False, opt_dirs=None):
    cmds = {
        'lsdev': '/usr/sbin/lsdev',
        'lscfg': '/usr/sbin/lscfg',
        'fcinfo': '/usr/sbin/fcinfo',
        'ioscan': '/usr/bin/ioscan',
        'fcmsutil': '/opt/fcms/bin/fcmsutil',
    }
    return cmds.get(cmd, None)


@pytest.mark.parametrize(
    ("test_input", "expected"),
    [
        pytest.param(
            {
                "platform": "aix6",
                "mock_run_command": [(0, LSDEV_OUTPUT, ""), (0, LSCFG_OUTPUT, "")],
            },
            ["10000090FA551508"],
            id="aix6",
        ),
        pytest.param(
            {
                "platform": "sunos5",
                "mock_run_command": [
                    (0, FCINFO_OUTPUT, ""),
                ],
            },
            ["10000090fa1658de"],
            id="sunos5",
        ),
        pytest.param(
            {
                "platform": "hp-ux11",
                "mock_run_command": [(0, IOSCAN_OUT, ""), (0, FCMSUTIL_OUT, "")],
            },
            ["0x50060b00006975ec"],
            id="hp-ux11",
        ),
    ],
)
def test_get_fc_wwn_info(mocker, test_input, expected):
    module = mocker.MagicMock()
    inst = fc_wwn.FcWwnInitiatorFactCollector()

    mocker.patch("sys.platform", test_input["platform"])
    mocker.patch.object(module, "get_bin_path", side_effect=mock_get_bin_path)
    mocker.patch.object(
        module, "run_command", side_effect=test_input["mock_run_command"]
    )
    wwn_expected = {"fibre_channel_wwn": expected}
    assert wwn_expected == inst.collect(module=module)



# -*- coding: utf-8 -*-
# Copyright: (c) 2017, Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import annotations

import collections.abc as c
import os
import os.path
import pathlib
import re

import pytest

from ansible.config.manager import ConfigManager, ensure_type, resolve_path, get_config_type
from ansible.errors import AnsibleOptionsError, AnsibleError
from ansible._internal._datatag._tags import Origin, VaultedValue
from ansible.module_utils._internal._datatag import AnsibleTagHelper
from ansible.template import is_trusted_as_template
from units.mock.vault_helper import VaultTestHelper

curdir = os.path.dirname(__file__)
cfg_file = os.path.join(curdir, 'test.cfg')
cfg_file2 = os.path.join(curdir, 'test2.cfg')
cfg_file3 = os.path.join(curdir, 'test3.cfg')


class CustomMapping(c.Mapping):
    def __init__(self, values: c.Mapping) -> None:
        self._values = values

    def __getitem__(self, key, /):
        return self._values[key]

    def __len__(self):
        return len(self._values)

    def __iter__(self):
        return iter(self._values)


class Unhashable:
    def __eq__(self, other): ...


@pytest.mark.parametrize("value, value_type, expected_value", [
    (None, 'str', None),  # all types share a common short-circuit for None
    (Unhashable(), 'bool', False),
    ('y', 'bool', True),
    ('yes', 'bool', True),
    ('on', 'bool', True),
    ('1', 'bool', True),
    ('true', 'bool', True),
    ('t', 'bool', True),
    (1, 'bool', True),
    (1.0, 'bool', True),
    (True, 'bool', True),
    ('n', 'bool', False),
    ('no', 'bool', False),
    ('off', 'bool', False),
    ('0', 'bool', False),
    ('false', 'bool', False),
    ('f', 'bool', False),
    (0, 'bool', False),
    (0.0, 'bool', False),
    (False, 'bool', False),
    (False, 'boolean', False),  # alias
    ('10', 'int', 10),
    (20, 'int', 20),
    (True, 'int', 1),
    (False, 'int', 0),
    (42.0, 'int', 42),
    (-42.0, 'int', -42),
    (-42.0, 'integer', -42),  # alias
    ('2', 'float', 2.0),
    ('0.10', 'float', 0.10),
    (0.2, 'float', 0.2),
    ('a,b', 'list', ['a', 'b']),
    (['a', 1], 'list', ['a', 1]),
    (('a', 1), 'list', ['a', 1]),
    ('None', 'none', None),
    ('/p1', 'pathspec', ['/p1']),
    ('/p1:/p2', 'pathspec', ['/p1', '/p2']),
    ('/p1:/p2', 'pathspec', ['/p1', '/p2']),
    (['/p1', '/p2'], 'pathspec', ['/p1', '/p2']),
    ('/tmp/test.yml,/home/test2.yml', 'pathlist', ['/tmp/test.yml', '/home/test2.yml']),
    ('a', 'str', 'a'),
    ('Café', 'str', 'Café'),
    ('', 'str', ''),
    ('29', 'str', '29'),
    ('13.37', 'str', '13.37'),
    ('123j', 'str', '123j'),
    ('0x123', 'str', '0x123'),
    ('true', 'str', 'true'),
    ('True', 'str', 'True'),
    (0, 'str', '0'),
    (29, 'str', '29'),
    (13.37, 'str', '13.37'),
    (123j, 'str', '123j'),
    (0x123, 'str', '291'),
    (True, 'str', 'True'),
    (True, 'string', 'True'),  # alias
    (CustomMapping(dict(a=1)), 'dict', dict(a=1)),
    (dict(a=1), 'dict', dict(a=1)),
    (dict(a=1), 'dictionary', dict(a=1)),  # alias
    (123, 'bogustype', 123),  # unknown non-string types pass through unmodified
])
def test_ensure_type(value: object, value_type: str, expected_value: object) -> None:
    value = ensure_type(value, value_type)

    assert isinstance(value, type(expected_value))
    assert value == expected_value


@pytest.mark.parametrize("value, value_type, expected_msg_substring", [
    ('a', 'int', "Invalid value provided for 'int': 'a'"),
    ('NaN', 'int', "Invalid value provided for 'int': 'NaN'"),
    (b'10', 'int', "Invalid value provided for 'int': b'10'"),
    (1.1, 'int', "Invalid value provided for 'int': 1.1"),
    ('1.1', 'int', "Invalid value provided for 'int': '1.1'"),
    (-1.1, 'int', "Invalid value provided for 'int': -1.1"),
    ('a', 'float', "Invalid value provided for 'float': 'a'"),
    (b'a', 'float', "Invalid value provided for 'float': b'a'"),
    (1, 'list', "Invalid value provided for 'list': 1"),
    (b'a', 'list', "Invalid value provided for 'list': b'a'"),
    (1, 'none', "Invalid value provided for 'none': 1"),
    (1, 'path', "Invalid value provided for 'path': 1"),
    (1, 'tmp', "Invalid value provided for 'tmp': 1"),
    (1, 'pathspec', "Invalid value provided for 'pathspec': 1"),
    (b'a', 'pathspec', "Invalid value provided for 'pathspec': b'a'"),
    ([b'a'], 'pathspec', "Invalid value provided for 'pathspec': [b'a']"),
    (1, 'pathlist', "Invalid value provided for 'pathlist': 1"),
    (b'a', 'pathlist', "Invalid value provided for 'pathlist': b'a'"),
    ([b'a'], 'pathlist', "Invalid value provided for 'pathlist': [b'a']"),
    (1, 'dict', "Invalid value provided for 'dict': 1"),
    ([1], 'str', "Invalid value provided for 'str': [1]"),
])
def test_ensure_type_failure(value: object, value_type: str, expected_msg_substring: str) -> None:
    with pytest.raises(ValueError, match=re.escape(expected_msg_substring)):
        ensure_type(value, value_type)


@pytest.mark.parametrize("value, expected_value, value_type, origin, origin_ftype", [
    ('"value"', '"value"', 'str', 'env: ENVVAR', None),
    ('"value"', '"value"', 'str', os.path.join(curdir, 'test.yml'), 'yaml'),
    ('"value"', 'value', 'str', cfg_file, 'ini'),
    ('\'value\'', 'value', 'str', cfg_file, 'ini'),
    ('\'\'value\'\'', '\'value\'', 'str', cfg_file, 'ini'),
    ('""value""', '"value"', 'str', cfg_file, 'ini'),
    ('"x"', 'x', 'bogustype', cfg_file, 'ini'),  # unknown string types are unquoted
])
def test_ensure_type_unquoting(value: str, expected_value: str, value_type: str, origin: str | None, origin_ftype: str | None) -> None:
    actual_value = ensure_type(value, value_type, origin, origin_ftype)

    assert actual_value == expected_value


test_origin = Origin(description='abc')


@pytest.mark.parametrize("value, type", (
    (test_origin.tag('a,b,c'), 'list'),
    (test_origin.tag(('a', 'b')), 'list'),
    (test_origin.tag('1'), 'int'),
    (test_origin.tag('plainstr'), 'str'),
))
def test_ensure_type_tag_propagation(value: object, type: str) -> None:
    result = ensure_type(value, type)

    if value == result:
        assert value is result  # if the value wasn't transformed, it should be the same instance

    if isinstance(value, str) and isinstance(result, list):
        # split a str list; each value should be tagged
        assert all(Origin.is_tagged_on(v) for v in result)

    # the result should always be tagged
    assert Origin.is_tagged_on(result)


@pytest.mark.parametrize("value, type", (
    (test_origin.tag('plainstr'), 'tmp'),
))
def test_ensure_type_no_tag_propagation(value: object, type: str) -> None:
    result = ensure_type(value, type, origin='/tmp')

    assert not AnsibleTagHelper.tags(result)


@pytest.mark.parametrize("value, type", (
    ('blah1', 'temppath'),
    ('blah2', 'tmp'),
    ('blah3', 'tmppath'),
))
def test_ensure_type_temppath(value: object, type: str, tmp_path: pathlib.Path) -> None:
    path = ensure_type(value, type, origin=str(tmp_path))

    assert os.path.isdir(path)
    assert value in path
    assert os.listdir(path) == []


def test_ensure_type_vaulted(_vault_secrets_context: VaultTestHelper) -> None:
    raw = "secretvalue"
    origin = Origin(description='test')
    es = _vault_secrets_context.make_encrypted_string(raw)
    es = origin.tag(es)
    result = ensure_type(es, 'str')

    assert isinstance(result, str)
    assert result == raw
    assert VaultedValue.is_tagged_on(result)
    assert Origin.get_tag(result) is origin


class TestConfigManager:
    @classmethod
    def setup_class(cls):
        cls.manager = ConfigManager(cfg_file, os.path.join(curdir, 'test.yml'))

    @classmethod
    def teardown_class(cls):
        cls.manager = None

    def test_resolve_path(self):
        assert os.path.join(curdir, 'test.yml') == resolve_path('./test.yml', cfg_file)

    def test_resolve_path_cwd(self):
        assert os.path.join(os.getcwd(), 'test.yml') == resolve_path('{{CWD}}/test.yml')
        assert os.path.join(os.getcwd(), 'test.yml') == resolve_path('./test.yml')

    def test_value_and_origin_from_ini(self):
        assert self.manager.get_config_value_and_origin('config_entry') == ('fromini', cfg_file)

    def test_value_from_ini(self):
        assert self.manager.get_config_value('config_entry') == 'fromini'

    def test_value_and_origin_from_alt_ini(self):
        assert self.manager.get_config_value_and_origin('config_entry', cfile=cfg_file2) == ('fromini2', cfg_file2)

    def test_value_from_alt_ini(self):
        assert self.manager.get_config_value('config_entry', cfile=cfg_file2) == 'fromini2'

    def test_config_types(self):
        assert get_config_type('/tmp/ansible.ini') == 'ini'
        assert get_config_type('/tmp/ansible.cfg') == 'ini'
        assert get_config_type('/tmp/ansible.yaml') == 'yaml'
        assert get_config_type('/tmp/ansible.yml') == 'yaml'

    def test_config_types_negative(self):
        with pytest.raises(AnsibleOptionsError) as exec_info:
            get_config_type('/tmp/ansible.txt')
        assert "Unsupported configuration file extension for" in str(exec_info.value)

    def test_read_config_yaml_file(self):
        assert isinstance(self.manager._read_config_yaml_file(os.path.join(curdir, 'test.yml')), dict)

    def test_read_config_yaml_file_negative(self):
        with pytest.raises(AnsibleError) as exec_info:
            self.manager._read_config_yaml_file(os.path.join(curdir, 'test_non_existent.yml'))

        assert "Missing base YAML definition file (bad install?)" in str(exec_info.value)


@pytest.mark.parametrize(("key", "expected_value"), (
    ("COLOR_UNREACHABLE", "bright red"),
    ("COLOR_VERBOSE", "rgb013"),
    ("COLOR_DEBUG", "gray10")))
def test_256color_support(key, expected_value):
    # GIVEN: a config file containing 256-color values with default definitions
    manager = ConfigManager(cfg_file3)
    # WHEN: get config values
    actual_value = manager.get_config_value(key)
    # THEN: no error
    assert actual_value == expected_value


def test_config_trust_from_env(monkeypatch: pytest.MonkeyPatch) -> None:
    expected = "from test"
    monkeypatch.setenv("ANSIBLE_TEST_ENTRY", expected)
    result = ConfigManager().get_config_value("_Z_TEST_ENTRY")
    origin = Origin.get_tag(result)

    assert result == expected
    assert is_trusted_as_template(result)
    assert origin and origin.description == '<Config env: ANSIBLE_TEST_ENTRY>'


def test_config_trust_from_file(tmp_path: pathlib.Path) -> None:
    expected = "from test"
    cfg_path = tmp_path / 'test.cfg'

    cfg_path.write_text(f"[testing]\nvalid={expected}")

    result = ConfigManager(str(cfg_path)).get_config_value("_Z_TEST_ENTRY")
    origin = Origin.get_tag(result)

    assert result == expected
    assert is_trusted_as_template(result)
    assert origin
    assert origin.path == str(cfg_path)
    assert origin.description == "section 'testing' option 'valid'"



# (c) 2012-2014, Michael DeHaan <michael.dehaan@gmail.com>
#
# This file is part of Ansible
#
# Ansible is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Ansible is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Ansible.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

import errno
import io
import os
import signal
import sys
import textwrap
import traceback
import types
import typing as t

from multiprocessing.queues import Queue

from ansible._internal import _task
from ansible._internal._errors import _error_utils
from ansible.errors import AnsibleError
from ansible.executor.task_executor import TaskExecutor
from ansible.executor.task_queue_manager import FinalQueue, STDIN_FILENO, STDOUT_FILENO, STDERR_FILENO
from ansible.executor.task_result import _RawTaskResult
from ansible.inventory.host import Host
from ansible.module_utils.common.text.converters import to_text
from ansible.parsing.dataloader import DataLoader
from ansible.playbook.task import Task
from ansible.playbook.play_context import PlayContext
from ansible.utils.context_objects import CLIArgs
from ansible.utils.display import Display
from ansible.utils.multiprocessing import context as multiprocessing_context
from ansible.vars.manager import VariableManager

__all__ = ['WorkerProcess']

display = Display()

current_worker = None


class WorkerQueue(Queue):
    """Queue that raises AnsibleError items on get()."""
    def get(self, *args, **kwargs):
        result = super(WorkerQueue, self).get(*args, **kwargs)
        if isinstance(result, AnsibleError):
            raise result
        return result


class WorkerProcess(multiprocessing_context.Process):  # type: ignore[name-defined]
    """
    The worker thread class, which uses TaskExecutor to run tasks
    read from a job queue and pushes results into a results queue
    for reading later.
    """

    def __init__(
            self,
            *,
            final_q: FinalQueue,
            task_vars: dict,
            host: Host,
            task: Task,
            play_context: PlayContext,
            loader: DataLoader,
            variable_manager: VariableManager,
            shared_loader_obj: types.SimpleNamespace,
            worker_id: int,
            cliargs: CLIArgs
    ) -> None:

        super(WorkerProcess, self).__init__()
        # takes a task queue manager as the sole param:
        self._final_q = final_q
        self._task_vars = task_vars
        self._host = host
        self._task = task
        self._play_context = play_context
        self._loader = loader
        self._variable_manager = variable_manager
        self._shared_loader_obj = shared_loader_obj

        # NOTE: this works due to fork, if switching to threads this should change to per thread storage of temp files
        # clear var to ensure we only delete files for this child
        self._loader._tempfiles = set()

        self.worker_queue = WorkerQueue(ctx=multiprocessing_context)
        self.worker_id = worker_id

        self._cliargs = cliargs

    def _term(self, signum, frame) -> None:
        """In child termination when notified by the parent"""
        signal.signal(signum, signal.SIG_DFL)

        try:
            os.killpg(self.pid, signum)
            os.kill(self.pid, signum)
        except OSError as e:
            if e.errno != errno.ESRCH:
                signame = signal.strsignal(signum)
                display.error(f'Unable to send {signame} to child[{self.pid}]: {e}')

        # fallthrough, if we are still here, just die
        os._exit(1)

    def start(self) -> None:
        """
        multiprocessing.Process replaces the worker's stdin with a new file
        but we wish to preserve it if it is connected to a terminal.
        Therefore dup a copy prior to calling the real start(),
        ensuring the descriptor is preserved somewhere in the new child, and
        make sure it is closed in the parent when start() completes.
        """

        # FUTURE: this lock can be removed once a more generalized pre-fork thread pause is in place
        with display._lock:
            super(WorkerProcess, self).start()

    def _hard_exit(self, e: str) -> t.NoReturn:
        """
        There is no safe exception to return to higher level code that does not
        risk an innocent try/except finding itself executing in the wrong
        process. All code executing above WorkerProcess.run() on the stack
        conceptually belongs to another program.
        """

        try:
            display.debug(u"WORKER HARD EXIT: %s" % to_text(e))
        except BaseException:
            # If the cause of the fault is OSError being generated by stdio,
            # attempting to log a debug message may trigger another OSError.
            # Try printing once then give up.
            pass

        os._exit(1)

    def _detach(self) -> None:
        """
        The intent here is to detach the child process from the inherited stdio fds,
        including /dev/tty. Children should use Display instead of direct interactions
        with stdio fds.
        """
        try:
            os.setsid()
            # Create new fds for stdin/stdout/stderr, but also capture python uses of sys.stdout/stderr
            for fds, mode in (
                    ((STDIN_FILENO,), os.O_RDWR | os.O_NONBLOCK),
                    ((STDOUT_FILENO, STDERR_FILENO), os.O_WRONLY),
            ):
                stdio = os.open(os.devnull, mode)
                for fd in fds:
                    os.dup2(stdio, fd)
                os.close(stdio)
            sys.stdout = io.StringIO()
            sys.stderr = io.StringIO()
            sys.stdin = os.fdopen(STDIN_FILENO, 'r', closefd=False)
            # Close stdin so we don't get hanging workers
            # We use sys.stdin.close() for places where sys.stdin is used,
            # to give better errors, and to prevent fd 0 reuse
            sys.stdin.close()
        except Exception as e:
            display.error(f'Could not detach from stdio: {e}')
            os._exit(1)

    def run(self) -> None:
        """
        Wrap _run() to ensure no possibility an errant exception can cause
        control to return to the StrategyBase task loop, or any other code
        higher in the stack.

        As multiprocessing in Python 2.x provides no protection, it is possible
        a try/except added in far-away code can cause a crashed child process
        to suddenly assume the role and prior state of its parent.
        """
        # Set the queue on Display so calls to Display.display are proxied over the queue
        display.set_queue(self._final_q)
        self._detach()
        # propagate signals
        signal.signal(signal.SIGINT, self._term)
        signal.signal(signal.SIGTERM, self._term)
        try:
            with _task.TaskContext(self._task):
                return self._run()
        except BaseException:
            self._hard_exit(traceback.format_exc())

    def _run(self) -> None:
        """
        Called when the process is started.  Pushes the result onto the
        results queue. We also remove the host from the blocked hosts list, to
        signify that they are ready for their next task.
        """

        global current_worker

        current_worker = self

        executor_result = TaskExecutor(
            self._host,
            self._task,
            self._task_vars,
            self._play_context,
            self._loader,
            self._shared_loader_obj,
            self._final_q,
            self._variable_manager,
        ).run()

        self._host.vars = dict()
        self._host.groups = []

        for name, stdio in (('stdout', sys.stdout), ('stderr', sys.stderr)):
            if data := stdio.getvalue():  # type: ignore[union-attr]
                display.warning(
                    (
                        f'WorkerProcess for [{self._host}/{self._task}] errantly sent data directly to {name} instead of using Display:\n'
                        f'{textwrap.indent(data[:256], "    ")}\n'
                    ),
                    formatted=True
                )

        try:
            self._final_q.send_task_result(_RawTaskResult(
                host=self._host,
                task=self._task,
                return_data=executor_result,
                task_fields=self._task.dump_attrs(),
            ))
        except Exception as ex:
            try:
                raise AnsibleError("Task result omitted due to queue send failure.") from ex
            except Exception as ex_wrapper:
                self._final_q.send_task_result(_RawTaskResult(
                    host=self._host,
                    task=self._task,
                    # ignore the real task result and don't allow result object contribution from the exception (in case the pickling error was related)
                    return_data=_error_utils.result_dict_from_exception(ex_wrapper),
                    task_fields={},  # The failure pickling may have been caused by the task attrs, omit for safety
                ))



"""Unit tests to provide coverage not easily obtained from integration tests."""

from __future__ import annotations

from ansible.module_utils.basic import _get_available_hash_algorithms


def test_unavailable_algorithm(mocker):
    """Simulate an available algorithm that isn't."""
    expected_algorithms = {'sha256', 'sha512'}  # guaranteed to be available

    mocker.patch('hashlib.algorithms_available', expected_algorithms | {'not_actually_available'})

    available_algorithms = _get_available_hash_algorithms()

    assert sorted(expected_algorithms) == sorted(available_algorithms)


def test_fips_mode(mocker):
    """Simulate running in FIPS mode on Python 2.7.9 or later."""
    expected_algorithms = {'sha256', 'sha512'}  # guaranteed to be available

    mocker.patch('hashlib.algorithms_available', expected_algorithms | {'md5'})
    mocker.patch('hashlib.md5').side_effect = ValueError()  # using md5 in FIPS mode raises a ValueError

    available_algorithms = _get_available_hash_algorithms()

    assert sorted(expected_algorithms) == sorted(available_algorithms)



# Copyright (c) 2024 Ansible Project
# Simplified BSD License (see licenses/simplified_bsd.txt or https://opensource.org/licenses/BSD-2-Clause)

"""Support code for exclusive use by the AnsiballZ wrapper."""

from __future__ import annotations

import importlib
import json
import runpy
import sys
import typing as t

from ansible.module_utils import basic
from ansible.module_utils._internal import _errors, _traceback, _messages, _ansiballz
from ansible.module_utils.common.json import get_module_encoder, Direction


def run_module(
    *,
    json_params: bytes,
    profile: str,
    module_fqn: str,
    modlib_path: str,
    extensions: dict[str, dict[str, object]],
    init_globals: dict[str, t.Any] | None = None,
) -> None:  # pragma: nocover
    """Used internally by the AnsiballZ wrapper to run an Ansible module."""
    try:
        for extension, args in extensions.items():
            # importing _ansiballz instead of _extensions avoids an unnecessary import when extensions are not in use
            extension_module = importlib.import_module(f'{_ansiballz.__name__}._extensions.{extension}')
            extension_module.run(args)

        _run_module(
            json_params=json_params,
            profile=profile,
            module_fqn=module_fqn,
            modlib_path=modlib_path,
            init_globals=init_globals,
        )
    except Exception as ex:  # not BaseException, since modules are expected to raise SystemExit
        _handle_exception(ex, profile)


def _run_module(
    *,
    json_params: bytes,
    profile: str,
    module_fqn: str,
    modlib_path: str,
    init_globals: dict[str, t.Any] | None = None,
) -> None:
    """Used internally by `_run_module` to run an Ansible module after coverage has been enabled (if applicable)."""
    basic._ANSIBLE_ARGS = json_params
    basic._ANSIBLE_PROFILE = profile

    init_globals = init_globals or {}
    init_globals.update(_module_fqn=module_fqn, _modlib_path=modlib_path)

    # Run the module. By importing it as '__main__', it executes as a script.
    runpy.run_module(mod_name=module_fqn, init_globals=init_globals, run_name='__main__', alter_sys=True)

    # An Ansible module must print its own results and exit. If execution reaches this point, that did not happen.
    raise RuntimeError('New-style module did not handle its own exit.')


def _handle_exception(exception: BaseException, profile: str) -> t.NoReturn:
    """Handle the given exception."""
    result = dict(
        failed=True,
        exception=_messages.ErrorSummary(
            event=_errors.EventFactory.from_exception(exception, _traceback.is_traceback_enabled(_traceback.TracebackEvent.ERROR)),
        ),
    )

    encoder = get_module_encoder(profile, Direction.MODULE_TO_CONTROLLER)

    print(json.dumps(result, cls=encoder))  # pylint: disable=ansible-bad-function

    sys.exit(1)  # pylint: disable=ansible-bad-function



"""Test the patching infrastructure."""

from __future__ import annotations

import importlib
import pkgutil
import sys
import typing as t

import pytest

from ansible.module_utils._internal import _patches
from ansible.module_utils.common._utils import get_all_subclasses

module_to_patch = sys.modules[__name__]


def import_all_patches() -> None:
    """Import all patch modules so they're discoverable via `concrete_patch_types`."""
    for module_info in pkgutil.iter_modules(_patches.__path__, f'{_patches.__name__}.'):
        importlib.import_module(module_info.name)


def get_patch_types() -> list[type[_patches.CallablePatch]]:
    """Return a sorted list of all imported patches."""
    return sorted(get_all_subclasses(_patches.CallablePatch), key=lambda item: item.__name__)


def get_patch_required_test_cases() -> list:
    """
    Return a list of test cases for checking if patches have been applied.
    If a patch is not needed for a given Python version it will be marked as xfail.
    """
    import_all_patches()

    xfail_patch_when: dict[type[_patches.CallablePatch], bool] = {
        # Example:
        # _patches._some_patch_module.SomePatchClass: sys.version_info >= (3, 13),
    }

    patches = sorted(get_all_subclasses(_patches.CallablePatch), key=lambda item: item.__name__)
    patches = [pytest.param(patch, marks=pytest.mark.xfail) if xfail_patch_when.get(patch) else patch for patch in patches]

    return patches


@pytest.mark.parametrize("patch", get_patch_required_test_cases())
def test_patch_required(patch: _patches.CallablePatch) -> None:
    """
    Verify the patch is actually required on the currently running Python version.
    It's possible a future Python version may no longer require the patch.
    If this test fails, verify the patch is not required before ignoring the failure for the affected Python versions.
    """
    assert patch.is_patched()


@pytest.mark.parametrize("patch", get_patch_types())
def test_reload_patch(patch: type[_patches.CallablePatch]) -> None:
    """Make sure that reloading the patch doesn't end up double patching."""
    original_patch = patch
    original_patch_module = sys.modules[original_patch.__module__]
    original_callable = original_patch.get_current_implementation()

    current_patch_module = importlib.reload(original_patch_module)

    current_patch = getattr(current_patch_module, original_patch.__name__)
    current_patch.patch()
    current_callable = current_patch.get_current_implementation()

    assert current_patch is not original_patch
    assert current_patch_module is original_patch_module
    assert current_callable is original_callable


@pytest.mark.parametrize("patch", get_patch_types())
def test_delete_and_import_patch(patch: type[_patches.CallablePatch]) -> None:
    """Make sure that deleting and importing the patch doesn't end up double patching."""
    original_patch = patch
    original_patch_module = sys.modules[original_patch.__module__]
    original_callable = original_patch.get_current_implementation()

    del sys.modules[original_patch_module.__name__]

    current_patch_module = importlib.import_module(original_patch_module.__name__)

    current_patch = getattr(current_patch_module, original_patch.__name__)
    current_patch.patch()
    current_callable = current_patch.get_current_implementation()

    assert current_patch is not original_patch
    assert current_patch_module is not original_patch_module
    assert current_callable is original_callable


def func_for_ineffective_patch() -> bool:
    return False


def test_ineffective_patch() -> None:
    """Verify an ineffective patch results in an error being raised."""
    class IneffectivePatch(_patches.CallablePatch):
        target_container: t.ClassVar = module_to_patch
        target_attribute = 'func_for_ineffective_patch'

        @classmethod
        def is_patch_needed(cls) -> bool:
            return True

        def __call__(self) -> bool:
            return True

    original_function = module_to_patch.func_for_ineffective_patch

    with pytest.raises(RuntimeError, match="Validation of '.*' failed after patching."):
        IneffectivePatch.patch()

    assert module_to_patch.func_for_ineffective_patch is original_function


def func_for_patching_is_idempotent() -> bool:
    return False


def test_patching_is_idempotent() -> None:
    """Verify that patching is idempotent."""
    class PatchingIsIdempotentPatch(_patches.CallablePatch):
        target_container: t.ClassVar = module_to_patch
        target_attribute = 'func_for_patching_is_idempotent'

        @classmethod
        def is_patch_needed(cls) -> bool:
            return module_to_patch.func_for_patching_is_idempotent() is False

        def __call__(self) -> bool:
            return True

    original_function = module_to_patch.func_for_patching_is_idempotent

    PatchingIsIdempotentPatch.patch()

    first_patched_function = module_to_patch.func_for_patching_is_idempotent

    PatchingIsIdempotentPatch.patch()

    second_patched_function = module_to_patch.func_for_patching_is_idempotent

    assert first_patched_function is not original_function
    assert first_patched_function is second_patched_function



from __future__ import annotations

import typing as t

import pytest

from ansible.errors import AnsibleUndefinedVariable
from ansible._internal._datatag._tags import TrustedAsTemplate
from ansible._internal._templating._engine import TemplateEngine

TRUST = TrustedAsTemplate()


@pytest.mark.parametrize("expr, expected", (
    ('["ok", bogus_var, "also ok"] | select("defined")', ["ok", "also ok"]),
    ('[{"a": "ok"}, {"a": bogus_var}, {"a": "also ok"}] | selectattr("a", "defined")', [{"a": "ok"}, {"a": "also ok"}]),
    ('["ok", bogus_var, "also ok"] | reject("undefined")', ["ok", "also ok"]),
    ('[{"a": "ok"}, {"a": bogus_var}, {"a": "also ok"}] | rejectattr("a", "undefined")', [{"a": "ok"}, {"a": "also ok"}]),
    ('[["ok"], [bogus_var], ["also ok"]] | map("select", "defined")', [["ok"], [], ["also ok"]]),
    ('"abc" is defined', True),
    ('"abc" is undefined', False),
    ('bogus_var is defined', False),
    ('bogus_var is undefined', True),
))
def test_wrapped_plugins_undefined_arg_passthru(expr: str, expected: t.Any) -> None:
    assert TemplateEngine().evaluate_expression(TRUST.tag(expr)) == expected


@pytest.mark.parametrize("expr", (
    "['abc'] | selectattr('a', 'equalto', bogus_var)",
    "['abc'] | selectattr(attr=bogus_var)",
    "['abc'] | select('sameas', bogus_var)",
    "['abc'] | reject('sameas', bogus_var)",
    "['abc'] | rejectattr('a', 'equalto', bogus_var)",
    "['abc'] | rejectattr(attr=bogus_var)",
    "['abc'] | map(bogus_var)",
    "['abc'] | map(default=bogus_var)",
    'bogus_var is string',
    'bogus_var is float',
    'bogus_var is integer',
    'bogus_var is number',
    'bogus_var is none',
    'bogus_var is boolean',
    'bogus_var is false',
    'bogus_var is true',
    'bogus_var is mapping',
    'bogus_var is sameas(123)',
    '123 is sameas(bogus_var)',
    'bogus_var is escaped',
))
def test_wrapped_plugins_undefined_arg_fail(expr: str) -> None:
    with pytest.raises(AnsibleUndefinedVariable):
        TemplateEngine().evaluate_expression(TRUST.tag(expr))



# -*- coding: utf-8 -*-
# Copyright 2019, Andrew Klychkov @Andersson007 <aaklychkov@mail.ru>
# Simplified BSD License (see licenses/simplified_bsd.txt or https://opensource.org/licenses/BSD-2-Clause)

from __future__ import annotations

import pytest

from ansible.module_utils import _internal
from ansible.module_utils.common.text.converters import jsonify


@pytest.mark.parametrize(
    'test_input,expected',
    [
        (1, '1'),
        ('string', '"string"'),
        ('くらとみ', '"\\u304f\\u3089\\u3068\\u307f"'),
        ('café', '"caf\\u00e9"'),
        pytest.param(b'string', '"string"', marks=pytest.mark.skipif(_internal.is_controller, reason="bytes not supported by controller")),
        (False, 'false'),
        pytest.param('string'.encode('utf-8'), '"string"', marks=pytest.mark.skipif(_internal.is_controller, reason="bytes not supported by controller")),
    ]
)
def test_jsonify(test_input, expected):
    """Test for jsonify()."""
    assert jsonify(test_input) == expected



# -*- coding: utf-8 -*-
# Copyright (c) 2020 Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import annotations

import pytest

from ansible.utils.display import Display
from ansible.module_utils import basic
from units.test_utils.controller.display import emits_warnings


def test_module_utils_warn() -> None:
    """Verify that `module_utils.basic.warn` on the controller is routed to `Display.warning`."""
    with emits_warnings(warning_pattern="hello"):
        basic.warn("hello")


def test_module_utils_error_as_warning() -> None:
    """Verify that `module_utils.basic.error_as_warning` on the controller is routed to `Display.error_as_warning`."""
    with emits_warnings(warning_pattern="hello.*world"):
        try:
            raise Exception("world")
        except Exception as ex:
            basic.error_as_warning("hello", ex)


def test_module_utils_deprecate() -> None:
    """Verify that `module_utils.basic.deprecate` on the controller is routed to `Display.deprecated`."""
    with emits_warnings(deprecation_pattern="hello"):
        basic.deprecate("hello", version='9999.9')


@pytest.fixture
def warning_message():
    warning_message = 'bad things will happen'
    expected_warning_message = '[WARNING]: {0}\n'.format(warning_message)
    return warning_message, expected_warning_message


def test_warning(capsys, mocker, warning_message):
    warning_message, expected_warning_message = warning_message

    mocker.patch('ansible.utils.color.ANSIBLE_COLOR', True)
    mocker.patch('ansible.utils.color.parsecolor', return_value=u'1;35')  # value for 'bright purple'

    d = Display()
    d._warns.clear()
    d.warning(warning_message)
    out, err = capsys.readouterr()
    assert d._warns == {expected_warning_message}
    assert err == '\x1b[1;35m{0}\x1b[0m\n'.format(expected_warning_message.rstrip('\n'))


def test_warning_no_color(capsys, mocker, warning_message):
    warning_message, expected_warning_message = warning_message

    mocker.patch('ansible.utils.color.ANSIBLE_COLOR', False)

    d = Display()
    d._warns.clear()
    d.warning(warning_message)
    out, err = capsys.readouterr()
    assert d._warns == {expected_warning_message}
    assert err == expected_warning_message



# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

DOCUMENTATION = r"""
module: validate
short_description: validate
description: validate
author: "validate (@validate)"
"""

EXAMPLES = r"""
"""

RETURN = r"""
"""



# This file overrides the builtin ansible.module_utils.ansible_release file
# to test that it can be overridden. Previously this was facts.py but caused issues
# with dependencies that may need to execute a module that makes use of facts
data = 'overridden ansible_release.py'



# coding: utf-8
# (c) 2015, Toshio Kuratomi <tkuratomi@ansible.com>
#
# This file is part of Ansible
#
# Ansible is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Ansible is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Ansible.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

from ansible.parsing.quoting import unquote

import pytest

UNQUOTE_DATA = (
    (u'1', u'1'),
    (u'\'1\'', u'1'),
    (u'"1"', u'1'),
    (u'"1 \'2\'"', u'1 \'2\''),
    (u'\'1 "2"\'', u'1 "2"'),
    (u'\'1 \'2\'\'', u'1 \'2\''),
    (u'"1\\"', u'"1\\"'),
    (u'\'1\\\'', u'\'1\\\''),
    (u'"1 \\"2\\" 3"', u'1 \\"2\\" 3'),
    (u'\'1 \\\'2\\\' 3\'', u'1 \\\'2\\\' 3'),
    (u'"', u'"'),
    (u'\'', u'\''),
    # Not entirely sure these are good but they match the current
    # behaviour
    (u'"1""2"', u'1""2'),
    (u'\'1\'\'2\'', u'1\'\'2'),
    (u'"1" 2 "3"', u'1" 2 "3'),
    (u'"1"\'2\'"3"', u'1"\'2\'"3'),
)


@pytest.mark.parametrize("quoted, expected", UNQUOTE_DATA)
def test_unquote(quoted, expected):
    assert unquote(quoted) == expected



from __future__ import annotations as _annotations

import collections.abc as _c
import typing as _t

_T_co = _t.TypeVar('_T_co', covariant=True)


class SequenceProxy(_c.Sequence[_T_co]):
    """A read-only sequence proxy."""

    # DTFIX5: needs unit test coverage

    __slots__ = ('__value',)

    def __init__(self, value: _c.Sequence[_T_co]) -> None:
        self.__value = value

    @_t.overload
    def __getitem__(self, index: int) -> _T_co: ...

    @_t.overload
    def __getitem__(self, index: slice) -> _c.Sequence[_T_co]: ...

    def __getitem__(self, index: int | slice) -> _T_co | _c.Sequence[_T_co]:
        if isinstance(index, slice):
            return self.__class__(self.__value[index])

        return self.__value[index]

    def __len__(self) -> int:
        return len(self.__value)

    def __contains__(self, item: object) -> bool:
        return item in self.__value

    def __iter__(self) -> _t.Iterator[_T_co]:
        yield from self.__value

    def __reversed__(self) -> _c.Iterator[_T_co]:
        return reversed(self.__value)

    def index(self, *args) -> int:
        return self.__value.index(*args)

    def count(self, value: object) -> int:
        return self.__value.count(value)



# (c) 2012-2014, Michael DeHaan <michael.dehaan@gmail.com>
# (c) 2020 Ansible Project
#
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import annotations

import pytest
import re

from pytest_mock import MockerFixture

from ansible import context
from ansible.errors import AnsibleError
from ansible.plugins.loader import become_loader, shell_loader


def test_sudo(mocker, parser, reset_cli_args):
    options = parser.parse_args([])
    context._init_global_context(options)

    sudo = become_loader.get('sudo')
    sh = shell_loader.get('sh')
    sh.executable = "/bin/bash"

    sudo.set_options(direct={
        'become_user': 'foo',
        'become_flags': '-n -s -H',
    })

    cmd = sudo.build_become_command('/bin/foo', sh)

    assert re.match(r"""sudo\s+-n -s -H\s+-u foo /bin/bash -c 'echo BECOME-SUCCESS-.+? ; /bin/foo'""", cmd), cmd

    sudo.set_options(direct={
        'become_user': 'foo',
        'become_flags': '-n -s -H',
        'become_pass': 'testpass',
    })

    cmd = sudo.build_become_command('/bin/foo', sh)
    assert re.match(r"""sudo\s+-s\s-H\s+-p "\[sudo via ansible, key=.+?\] password:" -u foo /bin/bash -c 'echo BECOME-SUCCESS-.+? ; /bin/foo'""", cmd), cmd

    sudo.set_options(direct={
        'become_user': 'foo',
        'become_flags': '-snH',
        'become_pass': 'testpass',
    })

    cmd = sudo.build_become_command('/bin/foo', sh)
    assert re.match(r"""sudo\s+-sH\s+-p "\[sudo via ansible, key=.+?\] password:" -u foo /bin/bash -c 'echo BECOME-SUCCESS-.+? ; /bin/foo'""", cmd), cmd

    sudo.set_options(direct={
        'become_user': 'foo',
        'become_flags': '--non-interactive -s -H',
        'become_pass': 'testpass',
    })

    cmd = sudo.build_become_command('/bin/foo', sh)
    assert re.match(r"""sudo\s+-s\s-H\s+-p "\[sudo via ansible, key=.+?\] password:" -u foo /bin/bash -c 'echo BECOME-SUCCESS-.+? ; /bin/foo'""", cmd), cmd

    sudo.set_options(direct={
        'become_user': 'foo',
        'become_flags': '--non-interactive -nC5 -s -H',
        'become_pass': 'testpass',
    })

    cmd = sudo.build_become_command('/bin/foo', sh)
    assert re.match(r"""sudo\s+-C5\s-s\s-H\s+-p "\[sudo via ansible, key=.+?\] password:" -u foo /bin/bash -c 'echo BECOME-SUCCESS-.+? ; /bin/foo'""", cmd), cmd


@pytest.mark.parametrize("del_attr_name, expected_error_pattern", (
    ("ECHO", ".*does not support become.*missing the 'ECHO' attribute"),  # BecomeBase
    ("CD", ".*does not support sudo chdir.*missing the 'CD' attribute"),  # sudo
))
def test_invalid_shell_plugin(del_attr_name: str, expected_error_pattern: str, mocker: MockerFixture) -> None:
    def badprop(_self):
        raise AttributeError(del_attr_name)

    sh = shell_loader.get('sh')
    mocker.patch.object(type(sh), del_attr_name, property(fget=badprop))

    sudo = become_loader.get('sudo')
    sudo.set_options(direct=dict(sudo_chdir='/'))

    with pytest.raises(AnsibleError, match=expected_error_pattern):
        sudo.build_become_command('/stuff', sh)


def test_no_flags() -> None:
    sudo = become_loader.get('sudo')
    sudo.set_options(direct=dict(become_pass='x', become_flags=''))

    result = sudo.build_become_command('/stuff', shell_loader.get('sh'))

    # ensure no flags were in the final command other than -p and -u
    assert re.search(r'''^sudo +-p "[^"]*" -u root '[^']*'$''', result)


def test_no_cmd() -> None:
    cmd = ''

    assert become_loader.get('sudo').build_become_command(cmd, shell_loader.get('sh')) is cmd



from __future__ import annotations

DOCUMENTATION = """
    vars: noop_vars_plugin
    short_description: Do NOT load host and group vars
    description: don't test loading host and group vars from a collection
    options:
      stage:
        default: all
        choices: ['all', 'inventory', 'task']
        type: str
        ini:
          - key: stage
            section: testns.testcol.noop_vars_plugin
        env:
          - name: ANSIBLE_VARS_PLUGIN_STAGE
    extends_documentation_fragment:
        - testns.testcol2.deprecation
"""

from ansible.plugins.vars import BaseVarsPlugin


class VarsModule(BaseVarsPlugin):

    def get_vars(self, loader, path, entities, cache=True):
        super(VarsModule, self).get_vars(loader, path, entities)
        return {'collection': 'yes', 'notreal': 'value'}



from __future__ import annotations

import pytest
import pytest_mock

from ansible.module_utils._internal import _traceback


@pytest.mark.parametrize("patched_parsed_args, event, expected", (
    (dict(_ansible_tracebacks_for=["error", "warning"]), _traceback.TracebackEvent.ERROR, True),  # included value
    (dict(_ansible_tracebacks_for=["error", "warning"]), _traceback.TracebackEvent.WARNING, True),  # included value
    (dict(_ansible_tracebacks_for=["error", "warning"]), _traceback.TracebackEvent.DEPRECATED, False),  # excluded value
    ({}, _traceback.TracebackEvent.ERROR, False),  # unspecified defaults to no tracebacks
    (dict(_ansible_tracebacks_for="bogus,values"), _traceback.TracebackEvent.ERROR, True),  # parse failure defaults to always enabled
    (None, _traceback.TracebackEvent.ERROR, True),  # fetch failure defaults to always enabled
), ids=str)
def test_default_module_traceback_config(
        patched_parsed_args: dict | None,
        event: _traceback.TracebackEvent,
        expected: bool,
        mocker: pytest_mock.MockerFixture
) -> None:
    """Validate MU traceback config behavior (including unconfigured/broken config fallbacks)."""
    from ansible.module_utils import basic

    mocker.patch.object(basic, '_PARSED_MODULE_ARGS', patched_parsed_args)

    # this should just be an importlib.reload() on _traceback, but that redeclares the enum type and breaks the world
    mocker.patch.object(_traceback, '_module_tracebacks_enabled_events', None)

    assert _traceback._is_module_traceback_enabled(event=event) is expected



# This file is part of Ansible
#
# Ansible is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Ansible is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Ansible.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

import re


def get_sysctl(module, prefixes):

    sysctl = dict()
    sysctl_cmd = module.get_bin_path('sysctl')
    if sysctl_cmd is not None:

        cmd = [sysctl_cmd]
        cmd.extend(prefixes)

        try:
            rc, out, err = module.run_command(cmd)
        except OSError as ex:
            module.error_as_warning('Unable to read sysctl.', exception=ex)
            rc = 1

        if rc == 0:
            key = ''
            value = ''
            for line in out.splitlines():
                if not line.strip():
                    continue

                if line.startswith(' '):
                    # handle multiline values, they will not have a starting key
                    # Add the newline back in so people can split on it to parse
                    # lines if they need to.
                    value += '\n' + line
                    continue

                if key:
                    sysctl[key] = value.strip()

                try:
                    (key, value) = re.split(r'\s?=\s?|: ', line, maxsplit=1)
                except Exception as ex:
                    module.error_as_warning(f'Unable to split sysctl line {line!r}.', exception=ex)

            if key:
                sysctl[key] = value.strip()

    return sysctl



# Copyright (c), Sviatoslav Sydorenko <ssydoren@redhat.com> 2018
# Simplified BSD License (see licenses/simplified_bsd.txt or https://opensource.org/licenses/BSD-2-Clause)
"""Collections ABC import shim.

Use `ansible.module_utils.six.moves.collections_abc` instead, which has been available since ansible-core 2.11.
This module exists only for backwards compatibility.
"""

from __future__ import annotations

# Although this was originally intended for internal use only, it has wide adoption in collections.
# This is due in part to sanity tests previously recommending its use over `collections` imports.
from ansible.module_utils.six.moves.collections_abc import (  # pylint: disable=unused-import
    MappingView,
    ItemsView,
    KeysView,
    ValuesView,
    Mapping, MutableMapping,
    Sequence, MutableSequence,
    Set, MutableSet,
    Container,
    Hashable,
    Sized,
    Callable,
    Iterable,
    Iterator,
)



"""Analyze integration test target code coverage."""

from __future__ import annotations

import collections.abc as c
import os
import typing as t

from .....io import (
    read_json_file,
    write_json_file,
)

from .....util import (
    ApplicationError,
    display,
)

from .. import (
    CoverageAnalyzeConfig,
)

TargetKey = t.TypeVar('TargetKey', int, tuple[int, int])
TFlexKey = t.TypeVar('TFlexKey', int, tuple[int, int], str)
NamedPoints = dict[str, dict[TargetKey, set[str]]]
IndexedPoints = dict[str, dict[TargetKey, set[int]]]
Arcs = dict[str, dict[tuple[int, int], set[int]]]
Lines = dict[str, dict[int, set[int]]]
TargetIndexes = dict[str, int]
TargetSetIndexes = dict[frozenset[int], int]


class CoverageAnalyzeTargetsConfig(CoverageAnalyzeConfig):
    """Configuration for the `coverage analyze targets` command."""


def make_report(target_indexes: TargetIndexes, arcs: Arcs, lines: Lines) -> dict[str, t.Any]:
    """Condense target indexes, arcs and lines into a compact report."""
    set_indexes: TargetSetIndexes = {}
    arc_refs = dict((path, dict((format_arc(arc), get_target_set_index(indexes, set_indexes)) for arc, indexes in data.items())) for path, data in arcs.items())
    line_refs = dict((path, dict((line, get_target_set_index(indexes, set_indexes)) for line, indexes in data.items())) for path, data in lines.items())

    report = dict(
        targets=[name for name, index in sorted(target_indexes.items(), key=lambda kvp: kvp[1])],
        target_sets=[sorted(data) for data, index in sorted(set_indexes.items(), key=lambda kvp: kvp[1])],
        arcs=arc_refs,
        lines=line_refs,
    )

    return report


def load_report(report: dict[str, t.Any]) -> tuple[list[str], Arcs, Lines]:
    """Extract target indexes, arcs and lines from an existing report."""
    try:
        target_indexes: list[str] = report['targets']
        target_sets: list[list[int]] = report['target_sets']
        arc_data: dict[str, dict[str, int]] = report['arcs']
        line_data: dict[str, dict[int, int]] = report['lines']
    except KeyError as ex:
        raise ApplicationError('Document is missing key "%s".' % ex.args) from None
    except TypeError:
        raise ApplicationError('Document is type "%s" instead of "dict".' % type(report).__name__) from None

    arcs = dict((path, dict((parse_arc(arc), set(target_sets[index])) for arc, index in data.items())) for path, data in arc_data.items())
    lines = dict((path, dict((int(line), set(target_sets[index])) for line, index in data.items())) for path, data in line_data.items())

    return target_indexes, arcs, lines


def read_report(path: str) -> tuple[list[str], Arcs, Lines]:
    """Read a JSON report from disk."""
    try:
        report = read_json_file(path)
    except Exception as ex:
        raise ApplicationError('File "%s" is not valid JSON: %s' % (path, ex)) from None

    try:
        return load_report(report)
    except ApplicationError as ex:
        raise ApplicationError('File "%s" is not an aggregated coverage data file. %s' % (path, ex)) from None


def write_report(args: CoverageAnalyzeTargetsConfig, report: dict[str, t.Any], path: str) -> None:
    """Write a JSON report to disk."""
    if args.explain:
        return

    write_json_file(path, report, formatted=False)

    display.info('Generated %d byte report with %d targets covering %d files.' % (
        os.path.getsize(path), len(report['targets']), len(set(report['arcs'].keys()) | set(report['lines'].keys())),
    ), verbosity=1)


def format_line(value: int) -> str:
    """Format line as a string."""
    return str(value)  # putting this in a function keeps both pylint and mypy happy


def format_arc(value: tuple[int, int]) -> str:
    """Format an arc tuple as a string."""
    return '%d:%d' % value


def parse_arc(value: str) -> tuple[int, int]:
    """Parse an arc string into a tuple."""
    first, last = tuple(map(int, value.split(':')))
    return first, last


def get_target_set_index(data: set[int], target_set_indexes: TargetSetIndexes) -> int:
    """Find or add the target set in the result set and return the target set index."""
    return target_set_indexes.setdefault(frozenset(data), len(target_set_indexes))


def get_target_index(name: str, target_indexes: TargetIndexes) -> int:
    """Find or add the target in the result set and return the target index."""
    return target_indexes.setdefault(name, len(target_indexes))


def expand_indexes(
    source_data: IndexedPoints,
    source_index: list[str],
    format_func: c.Callable[[TargetKey], TFlexKey],
) -> dict[str, dict[TFlexKey, set[str]]]:
    """Expand indexes from the source into target names for easier processing of the data (arcs or lines)."""
    combined_data: dict[str, dict[TFlexKey, set[str]]] = {}

    for covered_path, covered_points in source_data.items():
        combined_points = combined_data.setdefault(covered_path, {})

        for covered_point, covered_target_indexes in covered_points.items():
            combined_point = combined_points.setdefault(format_func(covered_point), set())

            for covered_target_index in covered_target_indexes:
                combined_point.add(source_index[covered_target_index])

    return combined_data


def generate_indexes(target_indexes: TargetIndexes, data: NamedPoints) -> IndexedPoints:
    """Return an indexed version of the given data (arcs or points)."""
    results: IndexedPoints = {}

    for path, points in data.items():
        result_points = results[path] = {}

        for point, target_names in points.items():
            result_point = result_points[point] = set()

            for target_name in target_names:
                result_point.add(get_target_index(target_name, target_indexes))

    return results



# (c) 2012-2014, Michael DeHaan <michael.dehaan@gmail.com>
# (c) 2020 Ansible Project
#
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import annotations

import re

import pytest

from ansible import context
from ansible.plugins.loader import become_loader, shell_loader


def test_su(mocker, parser, reset_cli_args):
    options = parser.parse_args([])
    context._init_global_context(options)

    su = become_loader.get('su')
    sh = shell_loader.get('sh')
    sh.executable = "/bin/bash"

    su.set_options(direct={
        'become_user': 'foo',
        'become_flags': '',
    })

    cmd = su.build_become_command('/bin/foo', sh)
    assert re.match(r"""su\s+foo -c '/bin/bash -c '"'"'echo BECOME-SUCCESS-.+?; /bin/foo'"'"''""", cmd)


def test_no_cmd() -> None:
    cmd = ''

    assert become_loader.get('su').build_become_command(cmd, shell_loader.get('sh')) is cmd


@pytest.mark.parametrize("prefix, prompt", (
    ("", "Password:"),
    (" ", "Password :"),
    ("\n", "Password ："),
    ("x", "Password ："),
    ("", "口令:"),
    (" ", "口令 :"),
    ("\n", "口令 ："),
    ("x", "口令 ："),
))
def test_check_password_prompt_success(prefix: str, prompt: str) -> None:
    become = become_loader.get('su')

    assert become.check_password_prompt((prefix + prompt).encode()) is True
    assert become.prompt == prompt


@pytest.mark.parametrize("data", (
    "Password",
    "Passwort",
    "Pass:",
))
def test_check_password_prompt_failure(data: str) -> None:
    become = become_loader.get('su')

    assert become.check_password_prompt(data.encode()) is False
    assert become.prompt == ''


def test_check_password_prompt_escaping(mocker) -> None:
    become = become_loader.get('su')

    mocker.patch.object(become, 'get_option', return_value=['(invalid regex'])

    assert become.check_password_prompt('(invalid regex:') is True



# -*- coding: utf-8 -*-

# Copyright: (c) 2016, Krzysztof Magosa <krzysztof@magosa.pl>
# Copyright: (c) 2017, Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import annotations


DOCUMENTATION = """
---
module: tempfile
version_added: "2.3"
short_description: Creates temporary files and directories
description:
  - The M(ansible.builtin.tempfile) module creates temporary files and directories. C(mktemp) command
    takes different parameters on various systems, this module helps to avoid troubles related to that.
    Files/directories created by module are accessible only by creator. In case you need to make them
    world-accessible you need to use M(ansible.builtin.file) module.
  - For Windows targets, use the M(ansible.windows.win_tempfile) module instead.
options:
  state:
    description:
      - Whether to create file or directory.
    type: str
    choices: [ directory, file ]
    default: file
  path:
    description:
      - Location where temporary file or directory should be created.
      - If path is not specified, the default system temporary directory will be used.
    type: path
  prefix:
    description:
      - Prefix of file/directory name created by module.
    type: str
    default: ansible.
  suffix:
    description:
      - Suffix of file/directory name created by module.
    type: str
    default: ""
extends_documentation_fragment: action_common_attributes
attributes:
    check_mode:
        support: none
    diff_mode:
        support: none
    platform:
        platforms: posix
seealso:
- module: ansible.builtin.file
- module: ansible.windows.win_tempfile
author:
  - Krzysztof Magosa (@krzysztof-magosa)
"""

EXAMPLES = """
- name: Create temporary build directory
  ansible.builtin.tempfile:
    state: directory
    suffix: build

- name: Create temporary file
  ansible.builtin.tempfile:
    state: file
    suffix: temp
  register: tempfile_1

- name: Create a temporary file with a specific prefix
  ansible.builtin.tempfile:
     state: file
     suffix: txt
     prefix: myfile_

- name: Use the registered var and the file module to remove the temporary file
  ansible.builtin.file:
    path: "{{ tempfile_1.path }}"
    state: absent
  when: tempfile_1.path is defined
"""

RETURN = """
path:
  description: Path to created file or directory.
  returned: success
  type: str
  sample: "/tmp/ansible.bMlvdk"
"""

from os import close
from tempfile import mkstemp, mkdtemp

from ansible.module_utils.basic import AnsibleModule
from ansible.module_utils.common.text.converters import to_native


def main():
    module = AnsibleModule(
        argument_spec=dict(
            state=dict(type='str', default='file', choices=['file', 'directory']),
            path=dict(type='path'),
            prefix=dict(type='str', default='ansible.'),
            suffix=dict(type='str', default=''),
        ),
    )

    try:
        if module.params['state'] == 'file':
            handle, path = mkstemp(
                prefix=module.params['prefix'],
                suffix=module.params['suffix'],
                dir=module.params['path'],
            )
            close(handle)
        else:
            path = mkdtemp(
                prefix=module.params['prefix'],
                suffix=module.params['suffix'],
                dir=module.params['path'],
            )

        module.exit_json(changed=True, path=path)
    except Exception as e:
        module.fail_json(msg=to_native(e))


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

# Copyright: Contributors to the Ansible project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
from __future__ import annotations

import pytest

from ansible.module_utils.facts.virtual.sysctl import VirtualSysctlDetectionMixin


class MockVirtualSysctl(VirtualSysctlDetectionMixin):
    def __init__(self, module):
        self.module = module


@pytest.mark.parametrize("expected_path", ["/usr/sbin/sysctl", "/sbin/sysctl"])
def test_detect_sysctl(mocker, expected_path):
    module = mocker.Mock()
    module.get_bin_path.return_value = expected_path
    mixin = MockVirtualSysctl(module=module)
    mixin.detect_sysctl()

    assert mixin.sysctl_path == expected_path


@pytest.mark.parametrize(
    ("virt_product", "expected_guest"),
    [
        pytest.param(
            "KVM",
            "kvm",
            id="KVM-all-caps",
        ),
        pytest.param(
            "kvm",
            "kvm",
            id="kvm",
        ),
        pytest.param(
            "Bochs",
            "kvm",
            id="Bochs",
        ),
        pytest.param(
            "SmartDC",
            "kvm",
            id="SmartDC",
        ),
        pytest.param(
            "VMware",
            "VMware",
            id="VMware",
        ),
        pytest.param(
            "VirtualBox",
            "virtualbox",
            id="VirtualBox",
        ),
        pytest.param(
            "HVM domU",
            "xen",
            id="Xen-HVM",
        ),
        pytest.param(
            "XenPVH",
            "xen",
            id="Xen-PVH",
        ),
        pytest.param(
            "XenPV",
            "xen",
            id="Xen-PV",
        ),
        pytest.param(
            "XenPVHVM",
            "xen",
            id="Xen-PVHVM",
        ),
        pytest.param(
            "Hyper-V",
            "Hyper-V",
            id="Hyper-V",
        ),
        pytest.param(
            "Parallels",
            "parallels",
            id="Parallels",
        ),
        pytest.param(
            "RHEV Hypervisor",
            "RHEV",
            id="RHEV",
        ),
        pytest.param(
            "1",
            "jails",
            id="Jails",
        ),
    ],
)
def test_detect_virt_product(mocker, virt_product, expected_guest):
    module = mocker.Mock()
    module.get_bin_path.return_value = "/usr/bin/sysctl"
    module.run_command.return_value = (0, virt_product, "")
    mixin = MockVirtualSysctl(module=module)
    guest_facts = mixin.detect_virt_product("security.jail.jailed")
    expected = {
        "virtualization_role": "guest",
        "virtualization_tech_guest": set([expected_guest]),
        "virtualization_tech_host": set(),
        "virtualization_type": expected_guest,
    }
    assert guest_facts == expected


@pytest.mark.parametrize(
    ("virt_product", "expected_guest"),
    [
        pytest.param(
            "QEMU",
            "kvm",
            id="QEMU",
        ),
        pytest.param(
            "OpenBSD",
            "vmm",
            id="OpenBSD-vmm",
        ),
    ],
)
def test_detect_virt_vendor(mocker, virt_product, expected_guest):
    module = mocker.Mock()
    module.get_bin_path.return_value = "/usr/bin/sysctl"
    module.run_command.return_value = (0, virt_product, "")
    mixin = MockVirtualSysctl(module=module)
    guest_facts = mixin.detect_virt_vendor("security.jail.jailed")
    expected = {
        "virtualization_role": "guest",
        "virtualization_tech_guest": set([expected_guest]),
        "virtualization_tech_host": set(),
        "virtualization_type": expected_guest,
    }
    assert guest_facts == expected



from __future__ import annotations

from ansible.module_utils.basic import AnsibleModule


def main() -> None:
    module = AnsibleModule({})
    module.fail_json("the fail2 module went bang")


if __name__ == '__main__':
    main()
